#!/usr/bin/python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4
#
# Copyright © 2017 Philip Withnall
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for dbusapi.typeformatter
"""


from dbusapi.typeformatter import BasicTypeFormatter
from dbusapi import types
import unittest


class TestBasicFormatterNormal(unittest.TestCase):
    """Test normal formatting of types with the BasicTypeFormatter."""

    # pylint: disable=invalid-name
    def assertFormat(self, type_instance, formatted):  # noqa
        formatter = BasicTypeFormatter()
        self.assertEqual(formatter.format(type_instance), formatted)

    def test_basic_types(self):
        basic_type_map = {
            types.Byte: 'Byte',
            types.Boolean: 'Boolean',
            types.Int16: 'Int16',
            types.UInt16: 'UInt16',
            types.Int32: 'Int32',
            types.UInt32: 'UInt32',
            types.Int64: 'Int64',
            types.UInt64: 'UInt64',
            types.Double: 'Double',
            types.String: 'String',
            types.ObjectPath: 'Object Path',
            types.Signature: 'Signature',
            types.Variant: 'Variant',
            types.UnixFD: 'Unix FD',
        }

        for k, v in basic_type_map.items():
            self.assertFormat(k(), v)

    def test_array(self):
        type_instance = types.Array()
        type_instance.members.append(types.Byte())
        self.assertFormat(type_instance, 'Array of [Byte]')

    def test_struct(self):
        type_instance = types.Struct()
        type_instance.members.append(types.String())
        type_instance.members.append(types.String())
        type_instance.members.append(types.UInt16())
        self.assertFormat(type_instance, 'Struct of (String, String, UInt16)')

    def test_dict(self):
        dict_instance = types.DictEntry()
        dict_instance.members.append(types.String())
        dict_instance.members.append(types.Variant())
        array_instance = types.Array()
        array_instance.members.append(dict_instance)

        self.assertFormat(array_instance, 'Dict of {String: Variant}')

    def test_invalid_dict(self):
        """Test that a bare DictEntry (not in an Array) is not formatted"""
        dict_instance = types.DictEntry()
        dict_instance.members.append(types.String())
        dict_instance.members.append(types.Variant())

        formatter = BasicTypeFormatter()

        with self.assertRaises(AssertionError):
            formatter.format(dict_instance)

    def test_something_long(self):
        type_signature = types.TypeSignature()
        type_signature.members.append(types.Int32())
        type_signature.members.append(types.Int32())
        type_signature.members.append(types.Double())
        type_signature.members.append(types.Double())
        type_signature.members.append(types.Double())
        struct = types.Struct()
        struct.members.append(types.Int32())
        struct.members.append(types.Double())
        struct.members.append(types.Double())
        type_signature.members.append(struct)

        self.assertFormat(type_signature,
                          'Int32, Int32, Double, Double, Double, '
                          'Struct of (Int32, Double, Double)')

    def test_nested_structs(self):
        s1 = types.Struct()
        s1.members.append(types.Int32())
        s1.members.append(types.Int32())
        s2 = types.Struct()
        s2.members.append(types.UInt32())
        s2.members.append(types.UInt32())
        s3 = types.Struct()
        s3.members.append(s1)
        s3.members.append(s2)

        self.assertFormat(s3, 'Struct of ('
                              'Struct of (Int32, Int32), '
                              'Struct of (UInt32, UInt32))')

    def test_type_signature(self):
        signature = types.TypeSignature()
        signature.members.append(types.String())
        signature.members.append(types.UInt64())
        self.assertFormat(signature, 'String, UInt64')


if __name__ == '__main__':
    # Run test suite
    unittest.main()



# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4
#
# Copyright © 2016 Kaloyan Tenchov
# Copyright © 2017 Philip Withnall
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.


"""
Module providing a `TypeParser` object for parsing D-Bus type strings into
abstract syntax trees (ASTs) representing the nested type structure.
"""


from dbusapi import types
from dbusapi.log import Log


class TypeParsingLog(Log):

    """Specialized Log subclass for type parsing messages"""

    def __init__(self):
        """Construct a new TypeParsingLog"""
        super(TypeParsingLog, self).__init__()
        self.register_issue_code('invalid-type')
        self.register_issue_code('reserved-type')
        self.register_issue_code('unknown-type')
        self.domain = 'types'


class TypeParser(object):

    """
    Parse and validate a D-Bus type string.

    The D-Bus type system and type strings are explained in detail in the
    D-Bus specification:
    https://dbus.freedesktop.org/doc/dbus-specification.html#type-system
    and in the GVariant documentation for GLib:
    https://developer.gnome.org/glib/stable/glib-GVariantType.html#id-1.6.18.6.9
    """

    def __init__(self, signature):
        """
        Construct a new TypeParser.

        Args:
            signature: a D-Bus type string
        """
        self.signature = signature
        self._log = TypeParsingLog()
        self._index = 0

    @staticmethod
    def get_output_codes():
        """Return a list of all possible output codes."""
        return TypeParsingLog().issue_codes

    def get_output(self):
        """Return a list of all logged parser messages."""
        return self._log.issues

    def _get_next_character(self):
        """Return the next character from the signature."""
        if self._index < len(self.signature):
            character = self.signature[self._index]
            self._index += 1
        else:
            character = None
        return character

    # pylint: disable=too-many-return-statements,too-many-branches
    def _parse_one_type(self, character):
        """Parse one complete type from the signature."""
        basic_types = {
            'y': types.Byte,
            'b': types.Boolean,
            'n': types.Int16,
            'q': types.UInt16,
            'i': types.Int32,
            'u': types.UInt32,
            'x': types.Int64,
            't': types.UInt64,
            'd': types.Double,
            's': types.String,
            'o': types.ObjectPath,
            'g': types.Signature,
            'v': types.Variant,
            'h': types.UnixFD,
        }

        if character in basic_types:
            return basic_types[character]()
        elif character == "a":
            out_array = types.Array()
            character = self._get_next_character()
            if not character:
                self._log.log_issue('invalid-type',
                                    'Incomplete array declaration.')
                return None
            one_type = self._parse_one_type(character)
            if not one_type:
                # Invalid member type - error has already been logged.
                return None
            out_array.members.append(one_type)
            return out_array
        elif character == "(":
            out_struct = types.Struct()
            while True:
                character = self._get_next_character()
                if not character or \
                   (character == ')' and len(out_struct.members) == 0):
                    self._log.log_issue('invalid-type',
                                        'Incomplete structure declaration.')
                    return None
                elif character == ")":
                    break
                one_type = self._parse_one_type(character)
                if not one_type:
                    # Invalid member type - error has already been logged.
                    return None
                out_struct.members.append(one_type)
            return out_struct
        elif character == "{":
            out_dict = types.DictEntry()
            while True:
                character = self._get_next_character()
                if not character or \
                   (character == '}' and len(out_dict.members) != 2):
                    self._log.log_issue('invalid-type',
                                        'Incomplete dictionary declaration.')
                    return None
                elif character == "}":
                    break
                one_type = self._parse_one_type(character)
                if not one_type:
                    # Invalid member type - error has already been logged.
                    return None
                if len(out_dict.members) >= 2:
                    self._log.log_issue('invalid-type',
                                        'Invalid dictionary declaration.')
                    return None
                out_dict.members.append(one_type)
            return out_dict
        elif character in ['r', 'e', 'm', '*', '?', '@', '&', '^']:
            # https://dbus.freedesktop.org/doc/dbus-specification.html#idm399
            self._log.log_issue('reserved-type',
                                'Reserved type ‘%s’ must not be used in '
                                'signatures on D-Bus.' % character)
            return None
        else:
            self._log.log_issue('unknown-type',
                                'Unknown type ‘%s’.' % character)
            return None

    def parse(self):
        """
        Parse the type string and build an AST of the type

        Returns:
            A non-empty list of types.

            If parsing fails, or if the input string is empty,
            None is returned.
        """
        self._log.clear()
        self._index = 0

        if len(self.signature) == 0:
            self._log.log_issue('invalid-type', 'Empty type string.')

        out = types.TypeSignature()
        while True:
            character = self._get_next_character()
            if not character:
                break

            one_type = self._parse_one_type(character)
            if not one_type:
                # Invalid type
                break

            out.members.append(one_type)

        if self._log.issues:
            return None

        return out



#!/usr/bin/python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4
#
# Copyright © 2016 Kaloyan Tenchov
# Copyright © 2017 Philip Withnall
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for dbusapi.typeparser
"""


from dbusapi.typeparser import TypeParser
import unittest


def _test_parser(signature):
    """Build a TypeParser for a signature and parse it."""
    parser = TypeParser(signature)
    types = parser.parse()
    return parser, types


class TestParserErrors(unittest.TestCase):
    """Test error handling in the TypeParser."""

    # pylint: disable=invalid-name
    def assertOutput(self, signature, partial_output):  # noqa
        (parser, types) = _test_parser(signature)
        self.assertEqual(types, None)
        actual_output = \
            [(None, 'types', i[0], i[1]) for i in partial_output]
        self.assertEqual(parser.get_output(), actual_output)

    def test_reserved(self):
        """
        Type code 114 'r' is reserved for use in bindings and implementations
        to represent the general concept of a struct, and must not appear in
        signatures used on D-Bus.

        Similarly for the other type strings here.

        https://dbus.freedesktop.org/doc/dbus-specification.html#idm399
        """
        vectors = [
            'r', 'e', 'm', '*', '?', '@', '&', '^'
        ]
        for type_string in vectors:
            # Unfortunately we can’t use TestCase.subTest here until we depend
            # on Python 3.4 at a minimum.
            self.assertOutput(type_string, [
                ('reserved-type',
                 'Reserved type ‘%s’ must not be used in signatures on '
                 'D-Bus.' % type_string),
            ])

    def test_incomplete(self):
        self.assertOutput(
            "aa", [
                ('invalid-type', 'Incomplete array declaration.'),
            ])

    def test_incomplete2(self):
        self.assertOutput(
            "(ii", [
                ('invalid-type', 'Incomplete structure declaration.'),
            ])

    def test_incomplete3(self):
        self.assertOutput(
            "ii)", [
                ('unknown-type', 'Unknown type ‘)’.'),
            ])

    def test_incomplete4(self):
        self.assertOutput(
            "a{suu}", [
                ('invalid-type', 'Invalid dictionary declaration.'),
            ])

    def test_incomplete5(self):
        self.assertOutput(
            "a{su", [
                ('invalid-type', 'Incomplete dictionary declaration.'),
            ])

    def test_incomplete6(self):
        self.assertOutput(
            "a{s", [
                ('invalid-type', 'Incomplete dictionary declaration.'),
            ])

    def test_invalid_array(self):
        self.assertOutput(
            'a!', [
                ('unknown-type', 'Unknown type ‘!’.'),
            ]
        )

    def test_empty(self):
        self.assertOutput(
            "", [
                ('invalid-type', 'Empty type string.'),
            ]
        )

    def test_empty_tuple(self):
        self.assertOutput(
            '()', [
                ('invalid-type', 'Incomplete structure declaration.'),
            ]
        )

    def test_invalid_tuple(self):
        self.assertOutput(
            '(!', [
                ('unknown-type', 'Unknown type ‘!’.'),
            ]
        )

    def test_empty_dict(self):
        self.assertOutput(
            '{}', [
                ('invalid-type', 'Incomplete dictionary declaration.'),
            ]
        )

    def test_invalid_dict(self):
        self.assertOutput(
            '{!', [
                ('unknown-type', 'Unknown type ‘!’.'),
            ]
        )

    def test_underfilled_dict(self):
        self.assertOutput(
            '{s}', [
                ('invalid-type', 'Incomplete dictionary declaration.'),
            ]
        )

    def test_overfilled_dict(self):
        self.assertOutput(
            '{sss}', [
                ('invalid-type', 'Invalid dictionary declaration.'),
            ]
        )


class TestParserNormal(unittest.TestCase):
    """Test normal parsing of unusual input in the TypeParser."""

    # pylint: disable=invalid-name
    def assertParse(self, signature):  # noqa
        (parser, type_signature) = _test_parser(signature)
        self.assertEqual(parser.get_output(), [])
        actual_signature = str(type_signature)
        self.assertEqual(signature, actual_signature)

    def test_parse_valid(self):
        vectors = [
            'y', 'b', 'n', 'q', 'i', 'ii', 'aiai', 'u', 'uu', 'x', 't', 'd',
            's', 'o', 'g', 'v', 'h', 'ai', 'aai', 'au', 'av', '(iii)',
            '(ii)(ii)', '(i(ii))', '(v)', '(ius)', 'a(ii)', 'a{us}', 'a{us}i'
        ]
        for type_string in vectors:
            # Unfortunately we can’t use TestCase.subTest here until we depend
            # on Python 3.4 at a minimum.
            self.assertParse(type_string)


class TestParserOutputCodes(unittest.TestCase):
    """Test the output codes from TypeParser."""

    def test_unique(self):
        codes = TypeParser.get_output_codes()
        self.assertEqual(len(codes), len(set(codes)))

    def test_non_empty(self):
        codes = TypeParser.get_output_codes()
        self.assertNotEqual(codes, [])


if __name__ == '__main__':
    # Run test suite
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright © 2015 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for dbus-interface-diff
"""


# pylint: disable=missing-docstring


from dbusapi.interfaceparser import InterfaceParser
from dbusdeviation.interfacecomparator import InterfaceComparator
import os
import tempfile
import unittest


def _create_temp_xml_file(xml):
    """Create a temporary XML file with the given contents."""
    tmp_fd, tmp_name = tempfile.mkstemp(suffix='.xml', text=True)
    xml_fd = os.fdopen(tmp_fd, 'wt')
    xml_fd.write(xml)
    xml_fd.close()

    return tmp_name


# pylint: disable=too-many-public-methods
class TestComparatorErrors(unittest.TestCase):
    """Test log output from InterfaceComparator."""

    def _test_comparator(self, old_xml, new_xml):
        """Build an InterfaceComparator for the two parsed XML snippets."""
        old_tmpfile = _create_temp_xml_file(old_xml)
        new_tmpfile = _create_temp_xml_file(new_xml)

        old_parser = InterfaceParser(old_tmpfile)
        new_parser = InterfaceParser(new_tmpfile)

        old_interfaces = old_parser.parse()
        new_interfaces = new_parser.parse()

        os.unlink(new_tmpfile)
        os.unlink(old_tmpfile)

        self.assertEqual(old_parser.get_output(), [])
        self.assertEqual(new_parser.get_output(), [])

        self.assertNotEqual(old_interfaces, None)
        self.assertNotEqual(new_interfaces, None)

        return InterfaceComparator(old_interfaces, new_interfaces)

    # pylint: disable=invalid-name
    def assertOutput(self, old_xml, new_xml, output):  # noqa
        comparator = self._test_comparator(old_xml, new_xml)
        self.assertEqual(comparator.compare(), output)
        self.assertEqual(comparator.get_output(), output)

    def test_interface_removed(self):
        self.assertOutput(
            "<node><interface name='I.A'/></node>",
            "<node></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'interface-removed',
                 'Interface ‘I.A’ has been removed.'),
            ])

    def test_interface_added(self):
        self.assertOutput(
            "<node></node>",
            "<node><interface name='I.A'/></node>", [
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'interface-added',
                 'Interface ‘I.A’ has been added.'),
            ])

    def test_interface_deprecated(self):
        self.assertOutput(
            "<node><interface name='I.A'/></node>",
            "<node><interface name='I.A'>"
            "<annotation name='org.freedesktop.DBus.Deprecated' value='true'/>"
            "</interface></node>", [
                (None, InterfaceComparator.OUTPUT_INFO,
                 'deprecated',
                 'Node ‘I.A’ has been deprecated.'),
            ])

    def test_interface_deprecated_explicit(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<annotation name='org.freedesktop.DBus.Deprecated' "
            "value='false'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<annotation name='org.freedesktop.DBus.Deprecated' value='true'/>"
            "</interface></node>", [
                (None, InterfaceComparator.OUTPUT_INFO,
                 'deprecated',
                 'Node ‘I.A’ has been deprecated.'),
            ])

    def test_interface_undeprecated(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<annotation name='org.freedesktop.DBus.Deprecated' value='true'/>"
            "</interface></node>",
            "<node><interface name='I.A'/></node>", [
                (None, InterfaceComparator.OUTPUT_INFO,
                 'undeprecated',
                 'Node ‘I.A’ has been un-deprecated.'),
            ])

    def test_interface_undeprecated_explicit(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<annotation name='org.freedesktop.DBus.Deprecated' value='true'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<annotation name='org.freedesktop.DBus.Deprecated' "
            "value='false'/>"
            "</interface></node>", [
                (None, InterfaceComparator.OUTPUT_INFO,
                 'undeprecated',
                 'Node ‘I.A’ has been un-deprecated.'),
            ])

    def test_interface_name_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'/></node>",
            "<node><interface name='I.A2'/></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'interface-removed',
                 'Interface ‘I.A’ has been removed.'),
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'interface-added',
                 'Interface ‘I.A2’ has been added.')
            ])

    def test_method_removed(self):
        self.assertOutput(
            "<node><interface name='I.A'><method name='M'/>"
            "</interface></node>",
            "<node><interface name='I.A'/></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'method-removed',
                 'Method ‘I.A.M’ has been removed.'),
            ])

    def test_method_added(self):
        self.assertOutput(
            "<node><interface name='I.A'/></node>",
            "<node><interface name='I.A'><method name='M'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'method-added',
                 'Method ‘I.A.M’ has been added.'),
            ])

    def test_method_name_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'><method name='M'/>"
            "</interface></node>",
            "<node><interface name='I.A'><method name='M2'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'method-removed',
                 'Method ‘I.A.M’ has been removed.'),
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'method-added',
                 'Method ‘I.A.M2’ has been added.'),
            ])

    def test_method_arg_removed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='s' direction='in'/></method>"
            "</interface></node>",
            "<node><interface name='I.A'><method name='M'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-removed',
                 'Argument 0 of method ‘I.A.M’ has been removed.'),
            ])

    def test_method_arg_added(self):
        self.assertOutput(
            "<node><interface name='I.A'><method name='M'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='s' direction='in'/></method>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-added',
                 'Argument 0 of method ‘I.A.M’ has been added.'),
            ])

    def test_method_arg_name_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='s' name='A' direction='in'/>"
            "</method></interface></node>",
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='s' name='Z' direction='in'/>"
            "</method></interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_INFO,
                 'argument-name-changed',
                 'Argument 0 (‘A’) of method ‘I.A.M’ has changed name '
                 'from ‘A’ to ‘Z’.'),
            ])

    def test_method_arg_type_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='s' direction='in'/></method>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='b' direction='in'/></method>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-type-changed',
                 'Argument 0 of method ‘I.A.M’ has changed type from ‘s’ to '
                 '‘b’.'),
            ])

    def test_method_arg_direction_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='s' direction='in'/></method>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<method name='M'><arg type='s' direction='out'/></method>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-direction-changed-in-out',
                 'Argument 0 of method ‘I.A.M’ has changed direction '
                 'from ‘in’ to ‘out’.'),
            ])

    def test_method_c_symbol_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'><method name='M'>"
            "<annotation name='org.freedesktop.DBus.GLib.CSymbol' value='S1'/>"
            "</method></interface></node>",
            "<node><interface name='I.A'><method name='M'>"
            "<annotation name='org.freedesktop.DBus.GLib.CSymbol' value='S2'/>"
            "</method></interface></node>", [
                (None, InterfaceComparator.OUTPUT_INFO,
                 'c-symbol-changed',
                 'Node ‘I.A.M’ has changed its C symbol from ‘S1’ to ‘S2’.'),
            ])

    def test_method_no_reply_changed_to_false(self):
        self.assertOutput(
            "<node><interface name='I.A'><method name='M'>"
            "<annotation name='org.freedesktop.DBus.Method.NoReply' "
            "value='true'/>"
            "</method></interface></node>",
            "<node><interface name='I.A'><method name='M'>"
            "<annotation name='org.freedesktop.DBus.Method.NoReply' "
            "value='false'/>"
            "</method></interface></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'reply-added',
                 'Node ‘I.A.M’ has been marked as returning a reply.'),
            ])

    def test_method_no_reply_changed_to_true(self):
        self.assertOutput(
            "<node><interface name='I.A'><method name='M'/>"
            "</interface></node>",
            "<node><interface name='I.A'><method name='M'>"
            "<annotation name='org.freedesktop.DBus.Method.NoReply' "
            "value='true'/>"
            "</method></interface></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'reply-removed',
                 'Node ‘I.A.M’ has been marked as not returning a reply.'),
            ])

    def test_method_no_reply_changed_to_true_explicit(self):
        self.assertOutput(
            "<node><interface name='I.A'><method name='M'>"
            "<annotation name='org.freedesktop.DBus.Method.NoReply' "
            "value='false'/>"
            "</method></interface></node>",
            "<node><interface name='I.A'><method name='M'>"
            "<annotation name='org.freedesktop.DBus.Method.NoReply' "
            "value='true'/>"
            "</method></interface></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'reply-removed',
                 'Node ‘I.A.M’ has been marked as not returning a reply.'),
            ])

    def test_property_removed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            "<node><interface name='I.A'/></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'property-removed',
                 'Property ‘I.A.P’ has been removed.'),
            ])

    def test_property_added(self):
        self.assertOutput(
            "<node><interface name='I.A'/></node>",
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'property-added',
                 'Property ‘I.A.P’ has been added.'),
            ])

    def test_property_name_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<property name='P2' type='b' access='readwrite'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'property-removed',
                 'Property ‘I.A.P’ has been removed.'),
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'property-added',
                 'Property ‘I.A.P2’ has been added.'),
            ])

    def test_property_type_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<property name='P' type='s' access='readwrite'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'property-type-changed',
                 'Property ‘I.A.P’ has changed type from ‘b’ to ‘s’.'),
            ])

    def test_property_access_changed_r_to_rw(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='read'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'property-access-changed-read-readwrite',
                 'Property ‘I.A.P’ has changed access from ‘read’ to '
                 '‘readwrite’, becoming less restrictive.'),
            ])

    def test_property_access_changed_w_to_rw(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='write'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'property-access-changed-write-readwrite',
                 'Property ‘I.A.P’ has changed access from ‘write’ to '
                 '‘readwrite’, becoming less restrictive.'),
            ])

    def test_property_access_changed_r_to_w(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='read'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='write'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'property-access-changed-read-write',
                 'Property ‘I.A.P’ has changed access '
                 'from ‘read’ to ‘write’.'),
            ])

    def test_property_access_changed_rw_to_r(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='read'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'property-access-changed-readwrite-read',
                 'Property ‘I.A.P’ has changed access from ‘readwrite’ to '
                 '‘read’.'),
            ])

    def test_property_access_changed_rw_to_w(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='readwrite'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<property name='P' type='b' access='write'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'property-access-changed-readwrite-write',
                 'Property ‘I.A.P’ has changed access from ‘readwrite’ to '
                 '‘write’.'),
            ])

    def test_property_emits_changed_signal_changed(self):
        # All the classes of error we expect.
        error1 = (
            InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
            'Node ‘I.A.P’ started emitting '
            'org.freedesktop.DBus.Properties.PropertiesChanged.'
        )
        error2a = (
            InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
            'Node ‘I.A.P’ stopped emitting its new value in '
            'org.freedesktop.DBus.Properties.PropertiesChanged.'
        )
        error2b = (
            InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
            'Node ‘I.A.P’ started emitting its new value in '
            'org.freedesktop.DBus.Properties.PropertiesChanged.'
        )
        error3 = (
            InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
            'Node ‘I.A.P’ stopped emitting '
            'org.freedesktop.DBus.Properties.PropertiesChanged.'
        )
        error4a = (
            InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
            'Node ‘I.A.P’ stopped being a constant.'
        )
        error4b = (
            InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
            'Node ‘I.A.P’ became a constant.'
        )

        # 2D matrix of test vectors. Each row gives the old annotation value;
        # each column the new annotation value. Rows and columns are labelled
        # by the corresponding index in labels. e.g. We expect error2a when
        # changing annotation value from 'true' to 'invalidates'.
        labels = ['true', 'invalidates', 'const', 'false']
        vectors = [
            [None, error2a, error3, error3],
            [error2b, None, error3, error3],
            [error1, error1, None, error4a],
            [error1, error1, error4b, None],
        ]

        for i in range(0, len(vectors)):
            for j in range(0, len(vectors[0])):
                vector = vectors[i][j]
                if vector is not None:
                    vector = (
                        None,
                        vector[0],
                        'ecs-changed-%s-%s' % (labels[i], labels[j]),
                        vector[1],
                    )
                expected_errors = [vector] if vector is not None else []

                self.assertOutput(
                    "<node><interface name='I.A'>"
                    "<property name='P' type='b' access='readwrite'>"
                    "<annotation "
                    "name='org.freedesktop.DBus.Property.EmitsChangedSignal' "
                    "value='%s'/>"
                    "</property>"
                    "</interface></node>" % labels[i],
                    "<node><interface name='I.A'>"
                    "<property name='P' type='b' access='readwrite'>"
                    "<annotation "
                    "name='org.freedesktop.DBus.Property.EmitsChangedSignal' "
                    "value='%s'/>"
                    "</property>"
                    "</interface></node>" % labels[j],
                    expected_errors)

    def test_signal_removed(self):
        self.assertOutput(
            "<node><interface name='I.A'><signal name='S'/>"
            "</interface></node>",
            "<node><interface name='I.A'/></node>", [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'signal-removed',
                 'Signal ‘I.A.S’ has been removed.'),
            ])

    def test_signal_added(self):
        self.assertOutput(
            "<node><interface name='I.A'/></node>",
            "<node><interface name='I.A'><signal name='S'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'signal-added',
                 'Signal ‘I.A.S’ has been added.'),
            ])

    def test_signal_name_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'><signal name='S'/>"
            "</interface></node>",
            "<node><interface name='I.A'><signal name='S2'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'signal-removed',
                 'Signal ‘I.A.S’ has been removed.'),
                (None, InterfaceComparator.OUTPUT_FORWARDS_INCOMPATIBLE,
                 'signal-added',
                 'Signal ‘I.A.S2’ has been added.'),
            ])

    def test_signal_arg_removed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='s' direction='in'/></signal>"
            "</interface></node>",
            "<node><interface name='I.A'><signal name='S'/>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-removed',
                 'Argument 0 of signal ‘I.A.S’ has been removed.'),
            ])

    def test_signal_arg_added(self):
        self.assertOutput(
            "<node><interface name='I.A'><signal name='S'/>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='s' direction='in'/></signal>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-added',
                 'Argument 0 of signal ‘I.A.S’ has been added.'),
            ])

    def test_signal_arg_name_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='s' name='A' direction='in'/></signal>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='s' name='Z' direction='in'/></signal>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_INFO,
                 'argument-name-changed',
                 'Argument 0 (‘A’) of signal ‘I.A.S’ has changed name '
                 'from ‘A’ to ‘Z’.'),
            ])

    def test_signal_arg_type_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='s' direction='in'/></signal>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='b' direction='in'/></signal>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-type-changed',
                 'Argument 0 of signal ‘I.A.S’ has changed type from ‘s’ to '
                 '‘b’.'),
            ])

    def test_signal_arg_direction_changed(self):
        self.assertOutput(
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='s' direction='in'/></signal>"
            "</interface></node>",
            "<node><interface name='I.A'>"
            "<signal name='S'><arg type='s' direction='out'/></signal>"
            "</interface></node>",
            [
                (None, InterfaceComparator.OUTPUT_BACKWARDS_INCOMPATIBLE,
                 'argument-direction-changed-in-out',
                 'Argument 0 of signal ‘I.A.S’ has changed direction '
                 'from ‘in’ to ‘out’.'),
            ])


if __name__ == '__main__':
    # Run test suite
    unittest.main()



# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4
#
# Copyright © 2017 Philip Withnall
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.


"""
Module providing a `BasicTypeFormatter` object for formatting D-Bus type
abstract syntax trees (ASTs) into human-readable strings describing the type
structure.

These strings are intended to be easy to read for a technical reader, but
without having to memorise the single-character D-Bus types. They could be used
in documentation, for example.

Other type formatters may be added in future which format the type ASTs
differently.
"""


from dbusapi import types


class BasicTypeFormatter(object):
    """
    Format a D-Bus type AST as a string.

    Given a valid D-Bus type structure (see types.Type and
    types.TypeSignature), format it as a human-readable string. The format
    mirrors that used by D-Feet when introspecting D-Bus interfaces.

    In future, support may be added to this class for choosing the output
    language for the formatted strings.
    """

    def format(self, type_instance):
        """
        Format the type instance as a human-readable string.

        Note that the type instance must be a valid type structure. For
        example, this means that any types.Struct instance must have one or
        more child members; any types.DictEntry instance must have exactly two
        child members (key and value) and be the immediate child of a
        types.Array instance.

        See the D-Bus documentation for the full details of type system
        restrictions:
        https://dbus.freedesktop.org/doc/dbus-specification.html#type-system

        The type instance may be a types.Type instance or a types.TypeSignature
        instance.

        Args:
            type_instance: valid types.Type or types.TypeSignature instance
        """
        basic_type_map = {
            types.Byte: 'Byte',
            types.Boolean: 'Boolean',
            types.Int16: 'Int16',
            types.UInt16: 'UInt16',
            types.Int32: 'Int32',
            types.UInt32: 'UInt32',
            types.Int64: 'Int64',
            types.UInt64: 'UInt64',
            types.Double: 'Double',
            types.String: 'String',
            types.ObjectPath: 'Object Path',
            types.Signature: 'Signature',
            types.Variant: 'Variant',
            types.UnixFD: 'Unix FD',
        }

        # pylint: disable=unidiomatic-typecheck
        if isinstance(type_instance, types.TypeSignature):
            return ', '.join(map(self.format, type_instance.members))
        elif (isinstance(type_instance, types.Array) and
              isinstance(type_instance.members[0], types.DictEntry)):
            dictionary = type_instance.members[0]
            key = self.format(dictionary.members[0])
            value = self.format(dictionary.members[1])
            return 'Dict of {{{}: {}}}'.format(key, value)
        elif isinstance(type_instance, types.Array):
            member = self.format(type_instance.members[0])
            return 'Array of [{}]'.format(member)
        elif isinstance(type_instance, types.Struct):
            members = map(self.format, type_instance.members)
            return 'Struct of ({})'.format(', '.join(members))
        elif type(type_instance) in basic_type_map:
            return basic_type_map[type(type_instance)]
        else:
            assert False



# -*- coding: utf-8 -*-
#
# Copyright © 2015, 2016 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.


"""
An implementation of the abstract syntax tree (AST) for a D-Bus introspection
document, which fully describes a D-Bus API.

An AST can be built by parsing an XML file (using
`interfaceparser.InterfaceParser`) or by building the tree of objects manually.
"""


from abc import ABCMeta
from collections import OrderedDict
# pylint: disable=no-member
from lxml import etree
from re import match
from dbusapi.log import Log
from dbusapi.typeparser import TypeParser


TP_DTD = 'http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0'
FDO_DTD = 'http://www.freedesktop.org/dbus/1.0/doc.dtd'


class AstLog(Log):

    """Specialized Log subclass for AST messages"""

    def __init__(self):
        """Construct a new AstLog"""
        super(AstLog, self).__init__()
        self.register_issue_code('unknown-node')
        self.register_issue_code('missing-attribute')
        self.register_issue_code('duplicate-node')
        self.register_issue_code('duplicate-interface')
        self.register_issue_code('duplicate-method')
        self.register_issue_code('duplicate-signal')
        self.register_issue_code('duplicate-property')
        self.register_issue_code('node-name')
        self.register_issue_code('interface-name')
        self.register_issue_code('method-name')
        self.register_issue_code('signal-name')
        self.register_issue_code('property-type')
        self.register_issue_code('argument-type')
        self.domain = 'ast'


def ignore_node(node):
    """Decide whether to ignore the given node when parsing."""
    return node.tag[0] == '{'  # in a namespace


# pylint: disable=too-many-instance-attributes
class BaseNode(object):

    """Base class for all D-Bus AST nodes."""

    __metaclass__ = ABCMeta

    DOCSTRING_TAGS = ['{%s}docstring' % TP_DTD,
                      '{%s}doc' % FDO_DTD]

    required_attributes = ['name']
    optional_attributes = []

    def __init__(self, name, annotations=None, log=None):
        """Construct a new ast.BaseNode

        Args:
            name: str, the name of the node.
            annotations: potentially empty dict of annotations applied
                to the node, mapping annotation name to an `ast.Annotation`
                instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        self.name = name
        self.children = []
        self.parent = None
        self._comment = None
        self.log = log or AstLog()
        self.annotations = OrderedDict()
        self.line_number = -1
        self.comment_line_number = -1

        self._children_types = {
            'annotation': Annotation,
        }
        self._type_containers = {
            Annotation: self.annotations,
        }

        for annotation in (annotations or {}).values():
            self._add_child(annotation)

    @classmethod
    def from_xml(cls, node, comment, log, parent=None):
        """Return a new ast.BaseNode instance from an XML node."""
        attrs = {}

        valid = True
        for attr_name in cls.required_attributes:
            # Avoid redefining a builtin
            member_name = attr_name
            if member_name == 'type':
                member_name = 'type_'

            try:
                attrs[member_name] = node.attrib[attr_name]
            except KeyError:
                log.log_issue('missing-attribute',
                              'Missing required attribute ‘%s’ in %s.' %
                              (attr_name, node.tag))
                valid = False

        if not valid:
            return None

        for attr_name in cls.optional_attributes:
            attrs[attr_name] = node.attrib.get(attr_name)

        # FIXME: Hack for the fact that Node.name and Argument.name are not
        # actually required, but is the first attribute in the constructor, and
        # hence must be specified. This can be removed when we break API.
        if (cls == Node or cls == Argument) and 'name' not in attrs:
            attrs['name'] = None
        elif issubclass(cls, Callable) and 'args' not in attrs:
            attrs['args'] = []

        attrs['log'] = log
        res = cls(**attrs)
        res.line_number = node.sourceline
        if comment is not None:
            res.comment = comment.text
            # lxml reports the last source line for xml comments as
            # being the actual source line, fix this.
            # Also report line numbers starting from 1, consistent
            # with node.line_number
            res.comment_line_number = (comment.sourceline -
                                       len(res.comment.split('\n')) + 1)

        if parent:
            parent.add_child(res)

        res.parse_xml_children(node)
        return res

    def add_child(self, child):
        """Add a child to the node"""
        return self._add_child(child)

    def parse_xml_children(self, node):
        """Parse the XML node's children."""
        xml_comment = None
        for elem in node:
            if elem.tag == etree.Comment:
                xml_comment = elem
                continue

            elif elem.tag in BaseNode.DOCSTRING_TAGS:
                self.comment_line_number = elem.sourceline
                self.comment = elem.text
                continue

            elif ignore_node(elem):
                xml_comment = None
                continue

            try:
                ctype = self._children_types[elem.tag]
            except KeyError:
                xml_comment = None
                if isinstance(self, Node) and not self.pretty_name:
                    # Special handling for root nodes to allow more meaningful
                    #   error messages.
                    self.__log_issue('unknown-node',
                                     "Unknown node ‘%s’ in root." % elem.tag)
                else:
                    self.__log_issue('unknown-node',
                                     "Unknown node ‘%s’ in %s ‘%s’." %
                                     (elem.tag, type(self).__name__.lower(),
                                      self.pretty_name))
                continue

            ctype.from_xml(elem, xml_comment, parent=self,
                           log=self.log)
            xml_comment = None

    def walk(self):
        """Traverse this node's children in pre-order."""
        for child in self.children:
            yield child
            for grandchild in child.walk():
                yield grandchild

    # Backward compat
    def format_name(self):
        """Format this node's name as a human-readable string"""
        return self.pretty_name

    @property
    def comment(self):
        """
        Get the comment for this node.

        Returns:
            str: If the node was annotated with `org.gtk.GDBus.DocString`,
                the value of the annotation, otherwise one of:
                * A tp:docstring child node
                * A doc:doc child node
                * An XML comment immediately preceding the XML node.
                , whichever is seen last.
        """
        try:
            doc_annotation = self.annotations['org.gtk.GDBus.DocString']
            return doc_annotation.value
        except KeyError:
            return self._comment

    @comment.setter
    def comment(self, value):
        """Set the comment for this node."""
        self._comment = value

    @property
    def pretty_name(self):
        """Format the node's name as a human-readable string."""
        return self.name

    def __log_issue(self, code, message):
        self.log.log_issue(code, message)

    def _child_is_duplicate(self, child):
        return child.name in self._type_containers[type(child)]

    def _add_child(self, child):
        child.parent = self
        if self._child_is_duplicate(child):
            self.__log_issue('duplicate-%s' % type(child).__name__.lower(),
                             'Duplicate %s definition ‘%s’.' %
                             (type(child).__name__.lower(), child.pretty_name))
            return False

        self.children.append(child)
        container = self._type_containers[type(child)]
        if isinstance(container, list):
            container.append(child)
        else:
            container[child.name] = child
        return True


class Node(BaseNode):

    """
    AST representation of a <node> element.

    This represents the top level of a D-Bus API.
    """

    required_attributes = []
    optional_attributes = ['name']

    # pylint: disable=too-many-arguments
    def __init__(self, name=None, interfaces=None, nodes=None,
                 annotations=None, log=None):
        """
        Construct a new ast.Node.

        Args:
            name: node name; a non-empty string; The root <node> should either
                have no name or should have a name that is a valid absolute
                object path. Child <node> names must be valid relative paths.
            interfaces: potentially empty dict of interfaces in the node,
                mapping interface name to an `ast.Interface` instance
            nodes: potentially empty dict of properties in the node,
                mapping node name to an `ast.Node` instance
            annotations: potentially empty dict of annotations applied to the
                node, mapping annotation name to an `ast.Annotation` instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Node, self).__init__(name, annotations, log)
        self._children_types.update({'interface': Interface,
                                     'node': Node})

        self.interfaces = OrderedDict()
        self.nodes = OrderedDict()

        self._type_containers.update({Interface: self.interfaces,
                                      Node: self.nodes})

        for child in (interfaces or {}).values():
            self._add_child(child)
        for child in (nodes or {}).values():
            self._add_child(child)

    def _add_child(self, child):
        if isinstance(child, Node):
            if not child.name:
                self.log.log_issue('missing-attribute',
                                   'Missing required attribute ‘name’ in '
                                   'non-root node.')
            elif not Node.is_valid_relative_object_path(child.name):
                self.log.log_issue('node-name',
                                   'Non-root node name is not a relative '
                                   'object path ‘%s’.' % child.name)
        child.node = self
        return super(Node, self)._add_child(child)

    @staticmethod
    def is_valid_absolute_object_path(path):
        """
        Validate an absolute D-Bus object path.

        https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-marshaling-object-path

        Args:
            path: object path
        """
        return path == '/' or match(r'(/[A-Za-z0-9_]+)+', path) is not None

    @staticmethod
    def is_valid_relative_object_path(path):
        """
        Validate a relative D-Bus object path.

        https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-marshaling-object-path

        Args:
            path: object path
        """
        return match(r'[A-Za-z0-9_]+(/[A-Za-z0-9_]+)*', path) is not None


class Interface(BaseNode):

    """
    AST representation of an <interface> element.

    This represents the most commonly used node of a D-Bus API.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, name, methods=None, properties=None,
                 signals=None, annotations=None, log=None):
        """
        Construct a new ast.Interface.

        Args:
            name: interface name; a non-empty string
            methods: potentially empty dict of methods in the interface,
                mapping method name to an `ast.Method` instance
            properties: potentially empty dict of properties in the interface,
                mapping property name to an `ast.Property` instance
            signals: potentially empty dict of signals in the interface,
                mapping signal name to an `ast.Signal` instance
            annotations: potentially empty dict of annotations applied to the
                interface, mapping annotation name to an `ast.Annotation`
                instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Interface, self).__init__(name, annotations, log)

        if name and not Interface.is_valid_interface_name(name):
            self.log.log_issue('interface-name',
                               'Invalid interface name ‘%s’.' % name)

        self._children_types.update({'signal': Signal,
                                     'method': Method,
                                     'property': Property})

        self.methods = OrderedDict()
        self.signals = OrderedDict()
        self.properties = OrderedDict()

        self._type_containers.update({Method: self.methods,
                                      Signal: self.signals,
                                      Property: self.properties})

        for child in (methods or {}).values():
            self._add_child(child)
        for child in (signals or {}).values():
            self._add_child(child)
        for child in (properties or {}).values():
            self._add_child(child)

    def _add_child(self, child):
        child.interface = self
        return super(Interface, self)._add_child(child)

    @staticmethod
    def is_valid_interface_name(name):
        """
        Validate a D-Bus interface name.

        http://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-interface

        Args:
            name: interface name
        """
        return len(name) <= 255 and \
            match(r'[A-Za-z_][A-Za-z0-9_]*'
                  r'(\.[A-Za-z_][A-Za-z0-9_]*)+', name) is not None


def _dotted_name(elem):
    if elem.parent:
        return elem.parent.format_name() + '.' + elem.name
    return elem.name


class Property(BaseNode):

    """
    AST representation of a <property> element.

    This represents a readable or writable property of an interface.
    """

    ACCESS_READ = 'read'
    ACCESS_WRITE = 'write'
    ACCESS_READWRITE = 'readwrite'

    required_attributes = BaseNode.required_attributes + ['access', 'type']

    # pylint: disable=too-many-arguments
    def __init__(self, name, type_, access, annotations=None, log=None):
        """
        Construct a new ast.Property.

        Args:
            name: property name; a non-empty string, not including the parent
                interface name
            type_: type string for the property; see http://goo.gl/uCpa5A
            access: ACCESS_READ, ACCESS_WRITE, or ACCESS_READWRITE
            annotations: potentially empty dict of annotations applied to the
                property, mapping annotation name to an `ast.Annotation`
                instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Property, self).__init__(name, annotations, log)

        type_parser = TypeParser(type_)
        self.type = type_parser.parse()
        if self.type is None:
            message = type_parser.get_output()[0][3]
            self.log.log_issue('property-type',
                               'Error when parsing type ‘%s’ for property '
                               '‘%s’: %s' %
                               (type_, name, message))

        self.access = access
        self.interface = None

    @property
    def pretty_name(self):
        """Format the property's name as a human-readable string"""
        return _dotted_name(self)


class Callable(BaseNode):

    u"""
    AST representation of a callable element.

    This represents a ‘callable’, such as a method or a signal. All callables
    contain a list of in and out arguments.
    """

    def __init__(self, name, args, annotations=None, log=None):
        """
        Construct a new ast.Callable.

        Args:
            name: callable name; a non-empty string, not including the parent
                interface name
            args: potentially empty ordered list of ast.Arguments accepted and
                returned by the callable
            annotations: potentially empty dict of annotations applied to the
                callable, mapping annotation name to an ast.Annotation
                instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Callable, self).__init__(name, annotations, log)
        self.arguments = []
        self.interface = None
        self._children_types.update({'arg': Argument})
        self._type_containers.update({Argument: self.arguments})

        for arg in args:
            self._add_child(arg)

    def _child_is_duplicate(self, child):
        if isinstance(child, Argument):
            return False
        return super(Callable, self)._child_is_duplicate(child)

    @property
    def pretty_name(self):
        """Format the callable's name as a human-readable string"""
        return _dotted_name(self)

    @staticmethod
    def is_valid_name(name):
        """
        Validate a D-Bus member name.

        https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-member

        Args:
            name: callable name
        """
        return len(name) <= 255 and \
            match(r'[A-Za-z_][A-Za-z0-9_]*', name) is not None


class Method(Callable):

    """
    AST representation of a <method> element.

    This represents a callable method of an interface.
    """

    def __init__(self, name, args, annotations=None, log=None):
        """
        Construct a new ast.Method.

        Args:
            name: method name; a non-empty string, not including the parent
                interface name
            args: potentially empty ordered list of ast.Arguments accepted and
                returned by the method
            annotations: potentially empty dict of annotations applied to the
                method, mapping annotation name to an ast.Annotation
                instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Method, self).__init__(name, args, annotations, log)

        if name and not Callable.is_valid_name(name):
            self.log.log_issue('method-name',
                               'Invalid method name ‘%s’.' % name)


class Signal(Callable):

    """
    AST representation of a <signal> element.

    This represents an emittable signal on an interface.
    """

    def __init__(self, name, args, annotations=None, log=None):
        """
        Construct a new ast.Signal.

        Args:
            name: signal name; a non-empty string, not including the parent
                interface name
            args: potentially empty ordered list of ast.Arguments accepted and
                returned by the signal
            annotations: potentially empty dict of annotations applied to the
                signal, mapping annotation name to an ast.Annotation
                instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Signal, self).__init__(name, args, annotations, log)

        if name and not Callable.is_valid_name(name):
            self.log.log_issue('signal-name',
                               'Invalid signal name ‘%s’.' % name)


class Argument(BaseNode):

    """
    AST representation of an <arg> element.

    This represents an argument to an `ast.Signal` or `ast.Method`.
    """

    DIRECTION_IN = 'in'
    DIRECTION_OUT = 'out'

    required_attributes = ['type']
    optional_attributes = ['direction', 'name']

    # pylint: disable=too-many-arguments
    def __init__(self, name, direction, type_, annotations=None, log=None):
        """
        Construct a new ast.Argument.

        Args:
            name: argument name; may be empty
            direction: DIRECTION_IN or DIRECTION_OUT
            type_: type string for the argument; see http://goo.gl/uCpa5A
            annotations: potentially empty dict of annotations applied to the
                argument, mapping annotation name to an ast.Annotation
                instance
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Argument, self).__init__(name, annotations, log)

        type_parser = TypeParser(type_)
        self.type = type_parser.parse()
        if self.type is None:
            message = type_parser.get_output()[0][3]
            self.log.log_issue('argument-type',
                               'Error when parsing type ‘%s’ for argument '
                               '‘%s’: %s' %
                               (type_, name, message))

        self.direction = direction or Argument.DIRECTION_IN
        self._index = -1

    @property
    def pretty_name(self):
        """Format the argument's name as a human-readable string"""
        if self.index == -1 and self.name is None:
            res = 'unnamed'
        elif self.index == -1:
            res = '‘%s’' % self.name
        elif self.name is None:
            res = '%u' % self.index
        else:
            res = '%u (‘%s’)' % (self.index, self.name)

        if self.parent:
            parent_type = type(self.parent).__name__.lower()
            res += ' of %s ‘%s’' % (parent_type, self.parent.pretty_name)

        return res

    @property
    def index(self):
        """The index of this argument in its parent's list of arguments"""
        # Slight optimization, assumes arguments cannot be reparented
        if self._index != -1:
            return self._index
        if not self.parent:
            return -1
        else:
            self._index = self.parent.arguments.index(self)
            return self._index


class Annotation(BaseNode):

    """
    AST representation of an <annotation> element.

    This represents an arbitrary key-value metadata annotation attached to one
    of the nodes in an interface.
    The annotation name can be one of the well-known ones described at
    http://goo.gl/LgmNUe, or could be something else.
    """

    optional_attributes = ['value']

    def __init__(self, name, value=None, log=None):
        """
        Construct a new ast.Annotation.

        Args:
            name: annotation name; a non-empty string
            value: annotation value; any string is permitted
            log: subclass of `Log`, used to store log messages; can be None
        """
        super(Annotation, self).__init__(name, log=log)
        self.value = value

    @property
    def pretty_name(self):
        """Format the annotation's name as a human-readable string"""
        if not self.parent:
            return self.name
        return '%s of ‘%s’' % (self.name, self.parent.pretty_name)



# -*- coding: utf-8 -*-
#
# Copyright © 2015 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
Module providing a `InterfaceComparator` object for comparing two D-Bus APIs
(provided as abstract syntax trees from the introspection XML), to determine if
they differ in API-incompatible ways.
"""

from dbusapi import ast


# Warning categories.
WARNING_CATEGORIES = [
    'info',
    'backwards-compatibility',
    'forwards-compatibility',
]


class InterfaceComparator(object):

    """
    Compare two D-Bus interface descriptions and determine how they differ.

    Differences are given different severity levels, depending on whether they
    affect

    * nothing, and are purely decorative; for example, changing the name of a
      method argument
    * forwards compatibility, where code written against the new interface
      may not work against the old interface; for example, because it uses a
      newly added method
    * backwards compatibility, where code written against the old interface
      may not work against the new interface; for example, because it changes
      the type of a property
    """

    # Output severity levels.
    OUTPUT_INFO = 'info'
    OUTPUT_FORWARDS_INCOMPATIBLE = 'forwards-compatibility'
    OUTPUT_BACKWARDS_INCOMPATIBLE = 'backwards-compatibility'

    def __init__(self, old_interfaces, new_interfaces,
                 enabled_warnings=None, disabled_warnings=None,
                 new_filename=None):
        """
        Construct a new InterfaceComparator.

        Args:
            old_interfaces: non-empty dict of old interfaces, mapping
                interface name to an ast.Interface instance
            new_interfaces: non-empty dict of new interfaces, mapping
                interface name to an ast.Interface instance
            enabled_warnings: potentially empty list of warning categories and
                codes to enable
            disabled_warnings: potentially empty list of warning categories and
                codes to disable
            new_filename: path to the new D-Bus interface file,
                or None if unknown
        """
        self._old_interfaces = old_interfaces
        self._new_interfaces = new_interfaces
        self._new_filename = new_filename
        self._output = []

        if enabled_warnings is not None:
            self._enabled_warnings = enabled_warnings
        else:
            self._enabled_warnings = WARNING_CATEGORIES

        if disabled_warnings is not None:
            self._disabled_warnings = disabled_warnings
        else:
            self._disabled_warnings = []

    @staticmethod
    def get_output_codes():
        """Return a list of all possible output codes."""
        # FIXME: Hard-coded for the moment.
        return [
            'interface-added',
            'interface-removed',
            'deprecated',
            'undeprecated'
            'c-symbol-changed',
            'reply-added',
            'reply-removed',
            'ecs-changed-true-invalidates',
            'ecs-changed-true-false',
            'ecs-changed-true-const',
            'ecs-changed-invalidates-true',
            'ecs-changed-invalidates-false',
            'ecs-changed-invalidates-const',
            'ecs-changed-false-invalidates',
            'ecs-changed-false-true',
            'ecs-changed-false-const',
            'ecs-changed-const-invalidates',
            'ecs-changed-const-true',
            'ecs-changed-const-false',
            'method-added',
            'method-removed',
            'property-added',
            'property-removed',
            'signal-added',
            'signal-removed',
            'argument-added',
            'argument-removed',
            'property-type-changed',
            'property-access-changed-read-readwrite',
            'property-access-changed-read-write',
            'property-access-changed-write-read',
            'property-access-changed-write-readwrite',
            'property-access-changed-readwrite-read',
            'property-access-changed-readwrite-write',
            'argument-name-changed',
            'argument-type-changed',
            'argument-direction-changed-in-out',
            'argument-direction-changed-out-in',
        ]

    def _issue_output(self, level, code, message):
        """Append a message to the comparator output."""
        self._output.append((self._new_filename, level, code, message))

    def _warning_enabled(self, level, code):
        """Determine whether the given output level is enabled for output."""
        return ((level in self._enabled_warnings and
                 level not in self._disabled_warnings and
                 code not in self._disabled_warnings) or
                (code in self._enabled_warnings and
                 code not in self._disabled_warnings))

    def get_output(self):
        """
        Return all the log messages generated by the latest call to compare().

        Disabled warnings will not be returned.
        """
        out = []

        for (filename, level, code, message) in self._output:
            if not self._warning_enabled(level, code):
                continue

            out.append((filename, level, code, message))

        return out

    def compare(self):
        """
        Compare the two interfaces and store the results.

        Returns:
            The list of relevant warnings to output; an empty list otherwise.
            The return value is affected by the categories of enabled
            warnings.
        """
        self._output = []

        for (name, interface) in self._old_interfaces.items():
            # See if the old interface exists in the new file.
            if name not in self._new_interfaces:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'interface-removed',
                                   'Interface ‘%s’ has been removed.' % name)
            else:
                # Compare the two.
                self._compare_interfaces(interface, self._new_interfaces[name])

        for (name, interface) in self._new_interfaces.items():
            # See if the new interface exists in the old file.
            if name not in self._old_interfaces:
                self._issue_output(self.OUTPUT_FORWARDS_INCOMPATIBLE,
                                   'interface-added',
                                   'Interface ‘%s’ has been added.' % name)

        # Work out the exit status.
        return self.get_output()

    # pylint: disable=too-many-branches
    def _compare_annotations(self, old_node, new_node):  # noqa
        """Compare two ast.Annotation instances."""

        def _get_string_annotation(node, annotation_name, default):
            """
            Get an annotation value as a string.

            Reference: http://goo.gl/3EtdNf

            Returns:
                The value of the `annotation_name` annotation as a string, or
                `default` if no annotation exists by that name.
            """
            if annotation_name in node.annotations:
                return node.annotations[annotation_name].value
            return default

        def _get_bool_annotation(node, annotation_name, default):
            """
            Get an annotation value as a boolean.

            Reference: http://goo.gl/3EtdNf

            Returns:
                The value of the `annotation_name` annotation as a boolean, or
                `default` if no annotation exists by that name.
            """
            if annotation_name in node.annotations:
                return node.annotations[annotation_name].value == 'true'
            return default

        def _get_ecs_annotation(node):
            """
            Get the value of the EmitsChangedSignal annotation.

            Reference: http://goo.gl/3EtdNf

            Returns:
                The value of the
                `org.freedesktop.DBus.Property.EmitsChangedSignal` annotation,
                if it exists, or the default, calculated as per the
                specification.
            """
            name = 'org.freedesktop.DBus.Property.EmitsChangedSignal'

            if name in node.annotations:
                return node.annotations[name].value
            elif isinstance(node, ast.Property):
                assert node.interface is not None
                return _get_ecs_annotation(node.interface)
            else:
                return 'true'

        old_deprecated = \
            _get_bool_annotation(old_node,
                                 'org.freedesktop.DBus.Deprecated', False)
        new_deprecated = \
            _get_bool_annotation(new_node,
                                 'org.freedesktop.DBus.Deprecated', False)

        if old_deprecated and not new_deprecated:
            self._issue_output(self.OUTPUT_INFO, 'undeprecated',
                               'Node ‘%s’ has been un-deprecated.' %
                               old_node.format_name())
        elif not old_deprecated and new_deprecated:
            self._issue_output(self.OUTPUT_INFO, 'deprecated',
                               'Node ‘%s’ has been deprecated.' %
                               old_node.format_name())

        old_c_symbol = \
            _get_string_annotation(old_node,
                                   'org.freedesktop.DBus.GLib.CSymbol', '')
        new_c_symbol = \
            _get_string_annotation(new_node,
                                   'org.freedesktop.DBus.GLib.CSymbol', '')

        if old_c_symbol != new_c_symbol:
            self._issue_output(self.OUTPUT_INFO, 'c-symbol-changed',
                               'Node ‘%s’ has changed its C symbol from ‘%s’ '
                               'to ‘%s’.' %
                               (old_node.format_name(), old_c_symbol,
                                new_c_symbol))

        old_no_reply = \
            _get_bool_annotation(old_node,
                                 'org.freedesktop.DBus.Method.NoReply', False)
        new_no_reply = \
            _get_bool_annotation(new_node,
                                 'org.freedesktop.DBus.Method.NoReply', False)

        if old_no_reply and not new_no_reply:
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                               'reply-added',
                               'Node ‘%s’ has been marked as returning a '
                               'reply.' % old_node.format_name())
        elif not old_no_reply and new_no_reply:
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                               'reply-removed',
                               'Node ‘%s’ has been marked as not returning a '
                               'reply.' % old_node.format_name())

        old_ecs = _get_ecs_annotation(old_node)
        new_ecs = _get_ecs_annotation(new_node)
        output_code = 'ecs-changed-%s-%s' % (old_ecs, new_ecs)

        #                                 New
        #                   | true | invalidates | const | false
        #     | true        | xxxx | B2          | F3    | F3
        # Old | invalidates | B2   | xxxxxxxxxxx | F3    | F3
        #     | const       | B1   | B1          | xxxxx | B4
        #     | false       | B1   | B1          | F4    | xxxxx
        #
        # B = Backwards-compatible; F = Forwards-compatible
        # 1 = Started notifying
        # 2 = Property switched lists in PropertiesChanged
        # 3 = Stopped notifying
        # 4 = const semantics changed

        if old_ecs in ['true', 'invalidates'] and \
           new_ecs in ['false', 'const']:
            self._issue_output(self.OUTPUT_FORWARDS_INCOMPATIBLE, output_code,
                               'Node ‘%s’ stopped emitting '
                               'org.freedesktop.DBus.Properties.'
                               'PropertiesChanged.' %
                               old_node.format_name())
        elif (old_ecs in ['false', 'const'] and
              new_ecs in ['true', 'invalidates']):
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE, output_code,
                               'Node ‘%s’ started emitting '
                               'org.freedesktop.DBus.Properties.'
                               'PropertiesChanged.' %
                               old_node.format_name())
        elif old_ecs == 'true' and new_ecs == 'invalidates':
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE, output_code,
                               'Node ‘%s’ stopped emitting its new value in '
                               'org.freedesktop.DBus.Properties.'
                               'PropertiesChanged.' %
                               old_node.format_name())
        elif old_ecs == 'invalidates' and new_ecs == 'true':
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE, output_code,
                               'Node ‘%s’ started emitting its new value in '
                               'org.freedesktop.DBus.Properties.'
                               'PropertiesChanged.' %
                               old_node.format_name())
        elif old_ecs == 'const' and new_ecs == 'false':
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE, output_code,
                               'Node ‘%s’ stopped being a constant.' %
                               old_node.format_name())
        elif old_ecs == 'false' and new_ecs == 'const':
            self._issue_output(self.OUTPUT_FORWARDS_INCOMPATIBLE, output_code,
                               'Node ‘%s’ became a constant.' %
                               old_node.format_name())

    # pylint: disable=too-many-branches
    def _compare_interfaces(self, old_interface, new_interface):  # noqa
        """Compare two ast.Interface instances."""
        # Precondition of calling this method.
        assert old_interface.name == new_interface.name

        # Compare methods.
        for (name, method) in old_interface.methods.items():
            if name not in new_interface.methods:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'method-removed',
                                   'Method ‘%s’ has been removed.' %
                                   method.format_name())
            else:
                self._compare_methods(method, new_interface.methods[name])

        for (name, method) in new_interface.methods.items():
            if name not in old_interface.methods:
                self._issue_output(self.OUTPUT_FORWARDS_INCOMPATIBLE,
                                   'method-added',
                                   'Method ‘%s’ has been added.' %
                                   method.format_name())

        # Compare properties
        for (name, prop) in old_interface.properties.items():
            if name not in new_interface.properties:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'property-removed',
                                   'Property ‘%s’ has been removed.' %
                                   prop.format_name())
            else:
                self._compare_properties(prop, new_interface.properties[name])

        for (name, prop) in new_interface.properties.items():
            if name not in old_interface.properties:
                self._issue_output(self.OUTPUT_FORWARDS_INCOMPATIBLE,
                                   'property-added',
                                   'Property ‘%s’ has been added.' %
                                   prop.format_name())

        # Compare signals
        for (name, signal) in old_interface.signals.items():
            if name not in new_interface.signals:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'signal-removed',
                                   'Signal ‘%s’ has been removed.' %
                                   signal.format_name())
            else:
                self._compare_signals(signal,
                                      new_interface.signals[name])

        for (name, signal) in new_interface.signals.items():
            if name not in old_interface.signals:
                self._issue_output(self.OUTPUT_FORWARDS_INCOMPATIBLE,
                                   'signal-added',
                                   'Signal ‘%s’ has been added.' %
                                   signal.format_name())

        # Compare annotations
        self._compare_annotations(old_interface, new_interface)

    def _compare_methods(self, old_method, new_method):
        """Compare two ast.Method instances."""
        # Precondition of calling this method.
        assert old_method.name == new_method.name

        # Compare the argument lists.
        n_old_args = len(old_method.arguments)
        n_new_args = len(new_method.arguments)

        for i in range(max(n_old_args, n_new_args)):
            if i >= n_old_args:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'argument-added',
                                   'Argument %s '
                                   'has been added.' %
                                   new_method.arguments[i].format_name())
            elif i >= n_new_args:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'argument-removed',
                                   'Argument %s '
                                   'has been removed.' %
                                   old_method.arguments[i].format_name())
            else:
                self._compare_arguments(old_method.arguments[i],
                                        new_method.arguments[i])

        # Compare annotations
        self._compare_annotations(old_method, new_method)

    def _compare_properties(self, old_property, new_property):
        """Compare two ast.Property instances."""
        # Precondition of calling this method.
        assert old_property.name == new_property.name

        if old_property.type != new_property.type:
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                               'property-type-changed',
                               'Property ‘%s’ has changed type from ‘%s’ '
                               'to ‘%s’.' %
                               (old_property.format_name(),
                                old_property.type, new_property.type))

        error_code = 'property-access-changed-%s-%s' % \
                     (old_property.access, new_property.access)

        if (old_property.access == ast.Property.ACCESS_READ or
            old_property.access == ast.Property.ACCESS_WRITE) and \
           new_property.access == ast.Property.ACCESS_READWRITE:
            # Property has become less restrictive.
            self._issue_output(self.OUTPUT_FORWARDS_INCOMPATIBLE, error_code,
                               'Property ‘%s’ has changed access from '
                               '‘%s’ to ‘%s’, becoming less restrictive.' %
                               (old_property.format_name(),
                                old_property.access, new_property.access))
        elif old_property.access != new_property.access:
            # Access has changed incompatibly.
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE, error_code,
                               'Property ‘%s’ has changed access from '
                               '‘%s’ to ‘%s’.' %
                               (old_property.format_name(),
                                old_property.access, new_property.access))

        # Compare annotations
        self._compare_annotations(old_property, new_property)

    def _compare_signals(self, old_signal, new_signal):
        """Compare two ast.Signal instances."""
        # Precondition of calling this method.
        assert old_signal.name == new_signal.name

        # Compare the argument lists.
        n_old_args = len(old_signal.arguments)
        n_new_args = len(new_signal.arguments)

        for i in range(max(n_old_args, n_new_args)):
            if i >= n_old_args:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'argument-added',
                                   'Argument %s '
                                   'has been added.' %
                                   new_signal.arguments[i].format_name())
            elif i >= n_new_args:
                self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                                   'argument-removed',
                                   'Argument %s '
                                   'has been removed.' %
                                   old_signal.arguments[i].format_name())
            else:
                self._compare_arguments(old_signal.arguments[i],
                                        new_signal.arguments[i])

        # Compare annotations
        self._compare_annotations(old_signal, new_signal)

    def _compare_arguments(self, old_arg, new_arg):
        """Compare two ast.Argument instances."""
        if old_arg.name != new_arg.name:
            self._issue_output(self.OUTPUT_INFO,
                               'argument-name-changed',
                               'Argument %s has changed '
                               'name from ‘%s’ to ‘%s’.' %
                               (old_arg.pretty_name,
                                old_arg.name, new_arg.name))

        if old_arg.type != new_arg.type:
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                               'argument-type-changed',
                               'Argument %s has changed '
                               'type from ‘%s’ to ‘%s’.' %
                               (old_arg.pretty_name,
                                old_arg.type, new_arg.type))

        if old_arg.direction != new_arg.direction:
            self._issue_output(self.OUTPUT_BACKWARDS_INCOMPATIBLE,
                               'argument-direction-changed-%s-%s' %
                               (old_arg.direction, new_arg.direction),
                               'Argument %s has changed '
                               'direction from ‘%s’ to ‘%s’.' %
                               (old_arg.pretty_name,
                                old_arg.direction, new_arg.direction))

        # Compare annotations
        self._compare_annotations(old_arg, new_arg)



# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4
#
# Copyright © 2016 Kaloyan Tenchov
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.


"""
A representation of the D-Bus type system as a series of classes which can be
built into an abstract syntax tree (AST) for representing complex (nested)
types.

An AST can be built by parsing a D-Bus type signature (using
`typeparser.TypeParser`) or by building the tree of objects manually.
"""


from abc import ABCMeta, abstractmethod


# pylint: disable=too-few-public-methods
class Type(object):
    """
    An abstract class - AST representation of a D-Bus type.

    See http://dbus.freedesktop.org/doc/dbus-specification.html#type-system
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """Constructor."""
        self.type = "\0"
        self.name = "INVALID"
        self.alignment = 1

    def __str__(self):
        """Format the type as a human-readable string."""
        return self.type

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)


# pylint: disable=too-few-public-methods
class Byte(Type):
    """
    AST representation of the D-Bus BYTE type.

    8-bit unsigned integer.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "y"
        self.name = "BYTE"
        self.alignment = 1


# pylint: disable=too-few-public-methods
class Boolean(Type):
    """
    AST representation of the D-Bus BOOLEAN type.

    Boolean value, 0 is FALSE and 1 is TRUE. Everything else is invalid.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "b"
        self.name = "BOOLEAN"
        self.alignment = 4


# pylint: disable=too-few-public-methods
class Int16(Type):
    """
    AST representation of the D-Bus INT16 type.

    16-bit signed integer.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "n"
        self.name = "INT16"
        self.alignment = 2


# pylint: disable=too-few-public-methods
class UInt16(Type):
    """
    AST representation of the D-Bus UINT16 type.

    16-bit unsigned integer.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "q"
        self.name = "UINT16"
        self.alignment = 2


# pylint: disable=too-few-public-methods
class Int32(Type):
    """
    AST representation of the D-Bus INT32 type.

    32-bit signed integer.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "i"
        self.name = "INT32"
        self.alignment = 4


# pylint: disable=too-few-public-methods
class UInt32(Type):
    """
    AST representation of the D-Bus UINT32 type.

    32-bit unsigned integer.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "u"
        self.name = "UINT32"
        self.alignment = 4


# pylint: disable=too-few-public-methods
class Int64(Type):
    """
    AST representation of the D-Bus INT64 type.

    64-bit signed integer.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "x"
        self.name = "INT64"
        self.alignment = 8


# pylint: disable=too-few-public-methods
class UInt64(Type):
    """
    AST representation of the D-Bus UINT64 type.

    64-bit unsigned integer.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "t"
        self.name = "UINT64"
        self.alignment = 8


# pylint: disable=too-few-public-methods
class Double(Type):
    """
    AST representation of the D-Bus DOUBLE type.

    IEEE 754 double.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "d"
        self.name = "DOUBLE"
        self.alignment = 8


# pylint: disable=too-few-public-methods
class String(Type):
    """
    AST representation of the D-Bus STRING type.

    UTF-8 string (must be valid UTF-8). Must be nul terminated and contain no
    other nul bytes.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "s"
        self.name = "STRING"
        self.alignment = 4


# pylint: disable=too-few-public-methods
class ObjectPath(Type):
    """
    AST representation of the D-Bus OBJECT_PATH type.

    Name of an object instance.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "o"
        self.name = "OBJECT_PATH"
        self.alignment = 4


# pylint: disable=too-few-public-methods
class Signature(Type):
    """
    AST representation of the D-Bus SIGNATURE type.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "g"
        self.name = "SIGNATURE"
        self.alignment = 1


# pylint: disable=too-few-public-methods
class Variant(Type):
    """
    AST representation of the D-Bus VARIANT type.

    Variant type - the type of the value is part of the value itself.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "v"
        self.name = "VARIANT"
        self.alignment = 1


# pylint: disable=too-few-public-methods
class UnixFD(Type):
    """
    AST representation of the D-Bus UNIX_FD type.

    Unix file descriptor.
    """

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.type = "h"
        self.name = "UNIX_FD"
        self.alignment = 4


# pylint: disable=too-few-public-methods
class Container(Type):
    """
    An abstract class - AST representation of the D-Bus container type.
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """Constructor."""
        Type.__init__(self)
        self.members = []

    @abstractmethod
    def __str__(self):
        """Format the type as a human-readable string."""
        pass


# pylint: disable=too-few-public-methods
class Array(Container):
    """
    AST representation of the D-Bus ARRAY type.
    """

    def __init__(self):
        """Constructor."""
        Container.__init__(self)
        self.type = "a"
        self.name = "ARRAY"
        self.alignment = 4

    def __str__(self):
        """Format the type as a human-readable string."""
        assert len(self.members) == 1
        return "{}{}".format(self.type, self.members[0])


# pylint: disable=too-few-public-methods
class Struct(Container):
    """
    AST representation of the D-Bus STRUCT type.
    """

    def __init__(self):
        """Constructor."""
        Container.__init__(self)
        self.type = "r"
        self.name = "STRUCT"
        self.alignment = 8

    def __str__(self):
        """Format the type as a human-readable string."""
        return "({})".format("".join(map(str, self.members)))


# pylint: disable=too-few-public-methods
class DictEntry(Container):
    """
    AST representation of the D-Bus DICT_ENTRY type.
    """

    def __init__(self):
        """Constructor."""
        Container.__init__(self)
        self.type = "e"
        self.name = "DICT_ENTRY"
        self.alignment = 8

    def __str__(self):
        """Format the type as a human-readable string."""
        assert len(self.members) == 2
        return "{{{}{}}}".format(self.members[0], self.members[1])


# pylint: disable=too-few-public-methods
class TypeSignature(object):
    """
    AST representation of a D-Bus signature - an ordered list of one or more
    types.

    See http://dbus.freedesktop.org/doc/dbus-specification.html#type-system
    """

    def __init__(self):
        """Constructor."""
        self.members = []

    def __str__(self):
        """Format the type as a human-readable string."""
        return "".join(map(str, self.members))

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright © 2015 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for dbusapi.ast
"""


# pylint: disable=missing-docstring


from dbusapi import ast
from dbusapi.types import TypeSignature
import unittest


# pylint: disable=too-many-public-methods
class TestAstNames(unittest.TestCase):
    """Test AST node name generation."""

    def test_interface(self):
        iface = ast.Interface('Some.Interface')
        self.assertEqual(iface.format_name(), 'Some.Interface')

    def test_property(self):
        prop = ast.Property('AProperty', 's', ast.Property.ACCESS_READ)
        ast.Interface('Some.Interface', {}, {
            'AProperty': prop,
        })
        self.assertEqual(prop.format_name(), 'Some.Interface.AProperty')

    def test_property_unparented(self):
        prop = ast.Property('AProperty', 's', ast.Property.ACCESS_READ)
        self.assertEqual(prop.format_name(), 'AProperty')

    def test_method(self):
        method = ast.Method('AMethod', [])
        ast.Interface('Some.Interface', {
            'AMethod': method,
        })
        self.assertEqual(method.format_name(), 'Some.Interface.AMethod')

    def test_method_unparented(self):
        method = ast.Method('AMethod', [])
        self.assertEqual(method.format_name(), 'AMethod')

    def test_signal(self):
        signal = ast.Signal('SomeSignal', [])
        ast.Interface('Some.Interface', {}, {}, {
            'SomeSignal': signal,
        })
        self.assertEqual(signal.format_name(), 'Some.Interface.SomeSignal')

    def test_signal_unparented(self):
        signal = ast.Signal('SomeSignal', [])
        self.assertEqual(signal.format_name(), 'SomeSignal')

    def test_argument(self):
        arg = ast.Argument('self', ast.Argument.DIRECTION_IN, 's')
        ast.Method('ParentMethod', [arg])
        self.assertEqual(arg.format_name(),
                         '0 (‘self’) of method ‘ParentMethod’')

    def test_argument_unparented(self):
        arg = ast.Argument('self', ast.Argument.DIRECTION_IN, 's')
        self.assertEqual(arg.format_name(), '‘self’')

    def test_argument_unnamed(self):
        arg = ast.Argument(None, ast.Argument.DIRECTION_IN, 's')
        ast.Method('ParentMethod', [arg])
        self.assertEqual(arg.format_name(), '0 of method ‘ParentMethod’')

    # pylint: disable=invalid-name
    def test_argument_unnamed_unparented(self):
        arg = ast.Argument(None, ast.Argument.DIRECTION_IN, 's')
        self.assertEqual(arg.format_name(), 'unnamed')

    def test_annotation_interface(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        ast.Interface('Some.Interface', {}, {}, {}, {
            'SomeAnnotation': annotation,
        })
        self.assertEqual(annotation.format_name(),
                         'SomeAnnotation of ‘Some.Interface’')

    def test_annotation_property(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        prop = ast.Property('AProperty', 's', ast.Property.ACCESS_READ, {
            'SomeAnnotation': annotation,
        })
        ast.Interface('Some.Interface', {}, {
            'AProperty': prop,
        })
        self.assertEqual(annotation.format_name(),
                         'SomeAnnotation of ‘Some.Interface.AProperty’')

    def test_annotation_method(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        method = ast.Method('AMethod', [], {
            'SomeAnnotation': annotation,
        })
        ast.Interface('Some.Interface', {
            'AMethod': method,
        })
        self.assertEqual(annotation.format_name(),
                         'SomeAnnotation of ‘Some.Interface.AMethod’')

    def test_annotation_signal(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        signal = ast.Signal('ASignal', [], {
            'SomeAnnotation': annotation,
        })
        ast.Interface('Some.Interface', {}, {}, {
            'ASignal': signal,
        })
        self.assertEqual(annotation.format_name(),
                         'SomeAnnotation of ‘Some.Interface.ASignal’')

    def test_annotation_argument(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        arg = ast.Argument('Argument', ast.Argument.DIRECTION_IN, 's', {
            'SomeAnnotation': annotation,
        })
        method = ast.Method('AMethod', [arg])
        ast.Interface('Some.Interface', {
            'AMethod': method,
        })
        self.assertEqual(annotation.format_name(),
                         'SomeAnnotation of ‘0 (‘Argument’)'
                         ' of method ‘Some.Interface.AMethod’’')

    def test_annotation_unparented(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        self.assertEqual(annotation.format_name(), 'SomeAnnotation')


# pylint: disable=too-many-public-methods
class TestAstParenting(unittest.TestCase):
    """Test AST node parenting."""

    def test_property(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        self.assertEqual(annotation.parent, None)

        prop = ast.Property('AProperty', 's', ast.Property.ACCESS_READ, {
            'SomeAnnotation': annotation,
        })
        self.assertEqual(prop.interface, None)
        self.assertEqual(annotation.parent, prop)

        iface = ast.Interface('Some.Interface', {}, {
            'AProperty': prop,
        })
        self.assertEqual(prop.interface, iface)

    def test_method(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        self.assertEqual(annotation.parent, None)

        method = ast.Method('AMethod', [], {
            'SomeAnnotation': annotation,
        })
        self.assertEqual(method.interface, None)
        self.assertEqual(annotation.parent, method)

        iface = ast.Interface('Some.Interface', {
            'AMethod': method,
        })
        self.assertEqual(method.interface, iface)

    def test_signal(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        self.assertEqual(annotation.parent, None)

        signal = ast.Signal('ASignal', [], {
            'SomeAnnotation': annotation,
        })
        self.assertEqual(signal.interface, None)
        self.assertEqual(annotation.parent, signal)

        iface = ast.Interface('Some.Interface', {}, {
            'ASignal': signal,
        })
        self.assertEqual(signal.interface, iface)

    def test_argument(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        self.assertEqual(annotation.parent, None)

        arg = ast.Argument('SomeArgument', ast.Argument.DIRECTION_IN, 's', {
            'SomeAnnotation': annotation,
        })
        self.assertEqual(arg.parent, None)
        self.assertEqual(arg.index, -1)
        self.assertEqual(annotation.parent, arg)

        method = ast.Method('AMethod', [arg])
        self.assertEqual(arg.parent, method)
        self.assertEqual(arg.index, 0)


class TestAstTraversal(unittest.TestCase):
    """Test AST traversal."""

    def test_walk(self):
        annotation = ast.Annotation('SomeAnnotation', 'value')
        method = ast.Method('AMethod', [], {
            'SomeAnnotation': annotation,
        })
        iface = ast.Interface('Some.Interface', {
            'AMethod': method,
        })

        children = [node for node in iface.walk()]
        self.assertEquals(len(children), 2)
        self.assertEquals(children[0], method)
        self.assertEquals(children[1], annotation)


class TestAstSignatures(unittest.TestCase):
    """Test AST node signature parsing."""

    def test_property(self):
        prop = ast.Property('AProperty', 's', ast.Property.ACCESS_READ)
        self.assertIsInstance(prop.type, TypeSignature)
        self.assertEqual(str(prop.type), 's')

    def test_argument(self):
        arg = ast.Argument('self', ast.Argument.DIRECTION_IN, 's')
        self.assertIsInstance(arg.type, TypeSignature)
        self.assertEqual(str(arg.type), 's')


class TestAstLogging(unittest.TestCase):
    """Test error handling in AST"""

    def test_duplicate(self):
        method = ast.Method('AMethod', [])
        iface = ast.Interface('Some.Interface', {
            'AMethod': method,
        })
        self.assertListEqual(iface.log.issues, [])
        duplicate_method = ast.Method('AMethod', [])
        iface.add_child(duplicate_method)
        self.assertListEqual(
            iface.log.issues,
            [(None,
              'ast',
              'duplicate-method',
              'Duplicate method definition ‘Some.Interface.AMethod’.')])


if __name__ == '__main__':
    # Run test suite
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright © 2015, 2016 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.


"""
An implementation of a base logging class.
"""


class Log(object):

    """Base logging class."""

    def __init__(self):
        """Construct a new Log"""
        self.issues = []
        self.issue_codes = set()
        self.domain = 'default'

    def register_issue_code(self, code):
        """
        Register a new issue code.

        Duplicate codes will be silently ignored.

        Args:
            code: str, an issue code, for example `unknown-node`
        """
        self.issue_codes.add(code)

    def log_issue(self, code, message):
        """
        Log a new issue.

        Args:
            code: str, A registered code for that issue.
            message: str, A message describing the issue.
        """
        assert code in self.issue_codes
        self.issues.append(self._create_entry(code, message))

    # pylint: disable=no-self-use
    def _create_entry(self, code, message):
        return None, self.domain, code, message

    def clear(self):
        """Clear the issue list."""
        self.issues = []



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright © 2015, 2016 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
D-Bus interface comparator. This allows two D-Bus introspection XML files to be
compared for API compatibility, and any incompatibilities debugged.

Compatibility warnings are split into categories, separating forwards- and
backwards-compatibility in case one or other is not cared about by the
interface maintainer. See the documentation for InterfaceComparator for an
explanation of the two types of compatibility.
"""

import argparse
import os
import sys

from lxml import etree

from dbusapi.interfaceparser import InterfaceParser
from dbusdeviation.interfacecomparator import InterfaceComparator

# Warning categories.
WARNING_CATEGORIES = [
    'info',
    'backwards-compatibility',
    'forwards-compatibility',
    'parser',
]


def _parse_file(filename, parser):
    """
    Parse an XML interface file.

    Return a list of ast.Interfaces. If the file is empty, return an empty
    list. Print an error and exit if parsing fails.
    """
    try:
        interfaces = parser.parse()

        # Handle parse errors.
        if interfaces is None:
            _print_output(parser.get_output())
            sys.exit(1)

        return interfaces
    # pylint: disable=E1101
    except etree.XMLSyntaxError as err:
        # If the file is empty, treat it as a non-existent Interface. This
        # allows for diffs of added files.
        if os.path.getsize(filename) == 0:
            return {}
        else:
            sys.stderr.write('%s: error: %s\n' % (filename, err))
            sys.exit(1)


def _format_level(level, enable_colour=True, justified_length=0):
    """Format a warning level as a human-readable string."""
    colours = {
        'info': '\033[96;1m',
        'forwards-compatibility': '\033[35;1m',
        'backwards-compatibility': '\033[91;1m',
        'parser': '\033[91;1m',
    }
    formats = {
        'info': 'note',
        'forwards-compatibility': 'warn',
        'backwards-compatibility': 'error',
        'parser': 'error',
    }

    out = formats[level]

    if justified_length > 0:
        out = out.rjust(justified_length)
    if enable_colour:
        out = colours[level] + out + '\033[0m'

    return out


def _get_fd_for_level(level):
    """Get the output file descriptor to use for the output level."""
    if level == InterfaceComparator.OUTPUT_INFO:
        return sys.stdout
    return sys.stderr


def _print_output(output, include_uris=True, enable_colour=True):
    """
    Print all the log messages generated by the latest call to compare().

    The messages will be printed to stdout and/or stderr as appropriate.
    """
    # Justify the error codes.
    try:
        max_code_len = max([len(o[2]) for o in output])
        max_level_len = max([len(_format_level(o[1], False)) for o in output])
    except ValueError:
        max_code_len = 0
        max_level_len = 0

    for (filename, level, code, message) in output:
        formatted_level = _format_level(level, enable_colour, max_level_len)
        fd_for_level = _get_fd_for_level(level)
        errors_page = 'https://tecnocode.co.uk/dbus-deviation/errors.html'
        explanation_uri = '%s#%s' % (errors_page, code)
        formatted_code = code.rjust(max_code_len)

        if enable_colour:
            formatted_code = '\033[1m%s\033[0m' % formatted_code
            explanation_uri = '\033[90m%s\033[0m' % explanation_uri
        else:
            formatted_code = formatted_code

        if filename is None:
            output = '%s: %s: %s\n' % \
                     (formatted_level, formatted_code, message)
        else:
            output = '%s: %s: %s: %s\n' % \
                     (filename, formatted_level, formatted_code, message)

        fd_for_level.write(output)
        if include_uris:
            fd_for_level.write('   %s\n' % explanation_uri)


def _calculate_exit_status(args, output):
    """
    Build the exit status for the program.

    This will be non-zero if any errors were outputted, or if --fatal-warnings
    was passed and any warnings were outputted. Otherwise it will be zero.
    """
    outputted_errors = False
    outputted_warnings = False

    for (_, level, _, _) in output:
        if level in ['backwards-compatibility', 'parser']:
            outputted_errors = True
        elif level in ['forwards-compatibility']:
            outputted_warnings = True

    if outputted_errors or (args.fatal_warnings and outputted_warnings):
        return len(output)

    return 0


def main():
    """Main utility implementation."""
    codes = InterfaceComparator.get_output_codes()
    codes += InterfaceParser.get_output_codes()

    # Parse command line arguments.
    parser = argparse.ArgumentParser(
        description='Comparing D-Bus interface definitions')
    parser.add_argument('old_file', type=str, help='Old interface XML file')
    parser.add_argument('new_file', type=str, help='New interface XML file')
    parser.add_argument('--file-display-name', dest='file_display_name',
                        type=str,
                        help='Human-readable name for new interface XML file '
                             '(if --new-file is a pipe, for example)')
    parser.add_argument('--fatal-warnings', action='store_const', const=True,
                        default=False, help='Treat all warnings as fatal')
    parser.add_argument('--warnings', dest='warnings', metavar='CATEGORY,…',
                        type=str,
                        help='Warning categories (%s; %s)' %
                             (', '.join(WARNING_CATEGORIES),
                              ', '.join(codes)))

    args = parser.parse_args()

    if not args.old_file or not args.new_file:
        parser.print_help()
        sys.exit(1)

    if args.warnings is None or args.warnings == 'all':
        # Enable all warnings by default
        warnings_args = WARNING_CATEGORIES
    elif args.warnings == 'none':
        warnings_args = []
    else:
        warnings_args = args.warnings.split(',')

    for warning_arg in warnings_args:
        if warning_arg[:3] == 'no-':
            warning_arg = warning_arg[3:]
        if warning_arg not in WARNING_CATEGORIES and warning_arg not in codes:
            sys.stderr.write('%s: Unrecognized warning ‘%s’.\n' %
                             (sys.argv[0], warning_arg))
            parser.print_help()
            sys.exit(1)

    enabled_warnings = [arg for arg in warnings_args if arg[:3] != 'no-']
    disabled_warnings = [arg[3:] for arg in warnings_args if arg[:3] == 'no-']

    # Parse the two files.
    old_parser = InterfaceParser(args.old_file)
    new_parser = InterfaceParser(args.new_file)

    old_interfaces = _parse_file(args.old_file, old_parser)
    new_interfaces = _parse_file(args.new_file, new_parser)

    # Work out the human-readable name of the new XML filename.
    if args.file_display_name is not None:
        new_filename = args.file_display_name
    else:
        new_filename = args.new_file

    # Compare the interfaces.
    comparator = InterfaceComparator(old_interfaces, new_interfaces,
                                     enabled_warnings, disabled_warnings,
                                     new_filename)
    out = comparator.compare()
    _print_output(out)
    sys.exit(_calculate_exit_status(args, out))


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright © 2015 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for dbusapi.interfaceparser
"""


# pylint: disable=missing-docstring


from dbusapi.interfaceparser import InterfaceParser
import os
import tempfile
import unittest


def _create_temp_xml_file(xml):
    """Create a temporary XML file with the given contents."""
    tmp_fd, tmp_name = tempfile.mkstemp(suffix='.xml', text=True)
    xml_fd = os.fdopen(tmp_fd, 'wt')
    xml_fd.write(xml)
    xml_fd.close()

    return tmp_name


def _test_parser_with_nodes(xml):
    """Build an InterfaceParser for the XML snippet and parse it."""
    tmpfile = _create_temp_xml_file(xml)
    parser = InterfaceParser(tmpfile)
    root_node = parser.parse_with_nodes()
    os.unlink(tmpfile)
    return parser, root_node, tmpfile


def _test_parser(xml):
    """Build an InterfaceParser for the XML snippet and parse it."""
    parser, root_node, tmpfile = _test_parser_with_nodes(xml)
    interfaces = root_node.interfaces if root_node else None
    return parser, interfaces, tmpfile


# pylint: disable=too-many-public-methods
class TestParserErrors(unittest.TestCase):
    """Test error handling in the InterfaceParser."""

    # pylint: disable=invalid-name
    def assertOutputWithNodes(self, xml, partial_output):  # noqa
        (parser, root_node, filename) = _test_parser_with_nodes(xml)
        self.assertEqual(root_node, None)
        actual_output = \
            [(filename, 'parser', i[0], i[1]) for i in partial_output]
        self.assertEqual(parser.get_output(), actual_output)

    # pylint: disable=invalid-name
    def assertOutput(self, xml, partial_output):  # noqa
        (parser, interfaces, filename) = _test_parser(xml)
        self.assertEqual(interfaces, None)
        actual_output = \
            [(filename, 'parser', i[0], i[1]) for i in partial_output]
        self.assertEqual(parser.get_output(), actual_output)

    def test_unknown_root_node(self):
        self.assertOutput(
            "<notnode><irrelevant/></notnode>", [
                ('unknown-node', 'Unknown root node ‘notnode’.'),
            ])

    def test_nonabsolute_node_name(self):
        self.assertOutputWithNodes(
            "<node name='rel/N'></node>", [
                ('node-name',
                 'Root node name is not an absolute object path ‘rel/N’.'),
            ])

    def test_invalid_node_name(self):
        self.assertOutputWithNodes(
            "<node name='//'></node>", [
                ('node-name',
                 'Root node name is not an absolute object path ‘//’.'),
            ])

    def test_missing_node_name(self):
        self.assertOutputWithNodes(
            "<node><node /></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in non-root node.'),
            ])

    def test_nonrelative_node_name(self):
        self.assertOutputWithNodes(
            "<node><node name='/abs/N'/></node>", [
                ('node-name',
                 'Non-root node name is not a relative object path ‘/abs/N’.'),
            ])

    def test_duplicate_node(self):
        self.assertOutputWithNodes(
            "<node><node name='N'/><node name='N'/></node>", [
                ('duplicate-node',
                 'Duplicate node definition ‘N’.'),
            ])

    def test_invalid_interface_name(self):
        self.assertOutput(
            "<node><interface name='0'/></node>", [
                ('interface-name',
                 'Invalid interface name ‘0’.'),
            ])

    def test_duplicate_interface(self):
        self.assertOutput(
            "<node><interface name='I.I'/><interface name='I.I'/></node>", [
                ('duplicate-interface',
                 'Duplicate interface definition ‘I.I’.'),
            ])

    def test_unknown_interface_node(self):
        self.assertOutput(
            "<node><badnode/></node>", [
                ('unknown-node', 'Unknown node ‘badnode’ in root.'),
            ])

    def test_interface_missing_name(self):
        self.assertOutput(
            "<node><interface/></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in interface.'),
            ])

    def test_invalid_method(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<method name='0M'/>"
            "</interface></node>", [
                ('method-name',
                 'Invalid method name ‘0M’.'),
            ])

    def test_duplicate_method(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<method name='M'/><method name='M'/>"
            "</interface></node>", [
                ('duplicate-method',
                 'Duplicate method definition ‘I.I.M’.'),
            ])

    def test_invalid_signal(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<signal name='*S'/>"
            "</interface></node>", [
                ('signal-name',
                 'Invalid signal name ‘*S’.'),
            ])

    def test_duplicate_signal(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<signal name='S'/><signal name='S'/>"
            "</interface></node>", [
                ('duplicate-signal',
                 'Duplicate signal definition ‘I.I.S’.'),
            ])

    def test_invalid_signal_signature(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<signal name='S'><arg name='N' type='?'/></signal>"
            "</interface></node>", [
                ('argument-type',
                 'Error when parsing type ‘?’ for argument ‘N’: '
                 'Reserved type ‘?’ must not be used in signatures on D-Bus.'),
            ])

    def test_duplicate_property(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<property name='P' type='s' access='readwrite'/>"
            "<property name='P' type='s' access='readwrite'/>"
            "</interface></node>", [
                ('duplicate-property',
                 'Duplicate property definition ‘I.I.P’.'),
            ])

    def test_invalid_property_signature(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<property name='P' type='a?' access='readwrite'/>"
            "</interface></node>", [
                ('property-type',
                 'Error when parsing type ‘a?’ for property ‘P’: '
                 'Reserved type ‘?’ must not be used in signatures on D-Bus.'),
            ])

    def test_unknown_sub_interface_node(self):
        self.assertOutput(
            "<node><interface name='I.I'><badnode/></interface></node>", [
                ('unknown-node', 'Unknown node ‘badnode’ in interface ‘I.I’.'),
            ])

    def test_method_missing_name(self):
        self.assertOutput(
            "<node><interface name='I.I'><method/></interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in method.'),
            ])

    def test_unknown_method_node(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<method name='M'><badnode/></method>"
            "</interface></node>", [
                ('unknown-node', 'Unknown node ‘badnode’ in method ‘I.I.M’.'),
            ])

    def test_signal_missing_name(self):
        self.assertOutput(
            "<node><interface name='I.I'><signal/></interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in signal.'),
            ])

    def test_unknown_signal_node(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<signal name='S'><badnode/></signal>"
            "</interface></node>", [
                ('unknown-node', 'Unknown node ‘badnode’ in signal ‘I.I.S’.'),
            ])

    def test_property_missing_name(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<property type='s' access='readwrite'/>"
            "</interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in property.'),
            ])

    def test_property_missing_type(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<property name='P' access='readwrite'/>"
            "</interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘type’ in property.'),
            ])

    def test_property_missing_access(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<property name='P' type='s'/>"
            "</interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘access’ in property.'),
            ])

    def test_unknown_property_node(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<property name='P' type='s' access='readwrite'>"
            "<badnode/>"
            "</property>"
            "</interface></node>", [
                ('unknown-node',
                 'Unknown node ‘badnode’ in property ‘I.I.P’.'),
            ])

    def test_unknown_arg_node(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<method name='M'><arg type='s'><badnode/></arg></method>"
            "</interface></node>", [
                ('unknown-node',
                 'Unknown node ‘badnode’ in argument ‘0 of method ‘I.I.M’’.'),
            ])

    def test_arg_missing_type(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<method name='M'><arg/></method>"
            "</interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘type’ in arg.'),
            ])

    def test_annotation_missing_name(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<annotation value='V'/>"
            "</interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in annotation.'),
            ])

    def test_unknown_annotation_node(self):
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<annotation name='N' value='V'><badnode/></annotation>"
            "</interface></node>", [
                ('unknown-node',
                 'Unknown node ‘badnode’ in annotation ‘N of ‘I.I’’.'),
            ])


# pylint: disable=too-many-public-methods
class TestParserNormal(unittest.TestCase):
    """Test normal parsing of unusual input in the InterfaceParser."""

    # pylint: disable=invalid-name
    def assertParse(self, xml):  # noqa
        (parser, interfaces, _) = _test_parser(xml)
        self.assertEqual(parser.get_output(), [])
        return interfaces

    def test_tp_spec_root(self):
        """Test that specifications wrapped in tp:spec are parsed."""
        self.assertParse(
            "<tp:spec xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "'><node><interface name='I.I'/></node></tp:spec>")

    def test_ignored_namespaced_tags_interface(self):
        self.assertParse(
            "<node><ignored:spec xmlns:ignored='http://ignored.com'>"
            "<interface name='I.I'/></ignored:spec></node>")

    def test_ignored_namespaced_tags(self):
        self.assertParse(
            "<node>"
            "<interface name='I.I'>"
            "<ignored:spec xmlns:ignored='http://ignored.com'>"
            "</ignored:spec>"
            "</interface></node>")

    def test_doc_root(self):
        """Test that doc tags are ignored in the root."""
        self.assertParse(
            "<node xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "' xmlns:doc='"
            "http://www.freedesktop.org/dbus/1.0/doc.dtd"
            "'>"
            "<tp:docstring>Ignore me.</tp:docstring>"
            "<doc:doc>Ignore me.</doc:doc>"
            "</node>")

    def test_doc_interface(self):
        """Test that doc tags are ignored in interfaces."""
        self.assertParse(
            "<node xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "' xmlns:doc='"
            "http://www.freedesktop.org/dbus/1.0/doc.dtd"
            "'><interface name='I.I'>"
            "<tp:docstring>Ignore me.</tp:docstring>"
            "<doc:doc>Ignore me.</doc:doc>"
            "</interface></node>")

    def test_doc_method(self):
        """Test that doc tags are ignored in methods."""
        self.assertParse(
            "<node xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "' xmlns:doc='"
            "http://www.freedesktop.org/dbus/1.0/doc.dtd"
            "'><interface name='I.I'><method name='M'>"
            "<tp:docstring>Ignore me.</tp:docstring>"
            "<doc:doc>Ignore me.</doc:doc>"
            "</method></interface></node>")

    def test_doc_signal(self):
        """Test that doc tags are ignored in signals."""
        self.assertParse(
            "<node xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "' xmlns:doc='"
            "http://www.freedesktop.org/dbus/1.0/doc.dtd"
            "'><interface name='I.I'><signal name='S'>"
            "<tp:docstring>Ignore me.</tp:docstring>"
            "<doc:doc>Ignore me.</doc:doc>"
            "</signal></interface></node>")

    def test_doc_property(self):
        """Test that doc tags are ignored in properties."""
        self.assertParse(
            "<node xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "' xmlns:doc='"
            "http://www.freedesktop.org/dbus/1.0/doc.dtd'>"
            "<interface name='I.I'><property name='P' type='s' access='read'>"
            "<tp:docstring>Ignore me.</tp:docstring>"
            "<doc:doc>Ignore me.</doc:doc>"
            "</property></interface></node>")

    def test_doc_arg(self):
        """Test that doc tags are ignored in args."""
        self.assertParse(
            "<node xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "' xmlns:doc='"
            "http://www.freedesktop.org/dbus/1.0/doc.dtd"
            "'><interface name='I.I'><method name='M'><arg type='s'>"
            "<tp:docstring>Ignore me.</tp:docstring>"
            "<doc:doc>Ignore me.</doc:doc>"
            "</arg></method></interface></node>")

    def test_doc_annotation(self):
        """Test that doc tags are ignored in annotations."""
        self.assertParse(
            "<node xmlns:tp='"
            "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
            "' xmlns:doc='"
            "http://www.freedesktop.org/dbus/1.0/doc.dtd"
            "'><interface name='I.I'><annotation name='A' value='V'>"
            "<tp:docstring>Ignore me.</tp:docstring>"
            "<doc:doc>Ignore me.</doc:doc>"
            "</annotation></interface></node>")

    def test_doc_comments(self):
        """Test that xml comments are *not* ignored"""
        xml = ("<node xmlns:tp='"
               "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
               "' xmlns:doc='"
               "http://www.freedesktop.org/dbus/1.0/doc.dtd'>"
               "<!--"
               "Please consider me"
               "-->"
               "<interface name='I.I'>"
               "<!--"
               "Notice me too"
               "-->"
               "<method name='foo'>"
               "<!--"
               "And me!"
               "-->"
               "<arg name='bar' type='s'/>"
               "</method>"
               "</interface></node>")

        (parser, interfaces, _) = _test_parser(xml)
        interface = interfaces.get('I.I')
        self.assertIsNotNone(interface)
        self.assertEquals(interface.comment, "Please consider me")
        meth = interface.methods.get('foo')
        self.assertIsNotNone(meth)
        self.assertEquals(meth.comment, "Notice me too")
        self.assertEquals(len(meth.arguments), 1)
        arg = meth.arguments[0]
        self.assertEquals(arg.comment, "And me!")

    def test_line_numbers(self):
        """Test that line numbers are correctly computed"""
        xml = ("<node xmlns:tp='"
               "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
               "' xmlns:doc='"
               "http://www.freedesktop.org/dbus/1.0/doc.dtd'>\n"
               "<!--\n"
               "Please consider me\n"
               "-->\n"
               "<interface name='I.I'>\n"
               "<!--\n"
               "Notice me too\n"
               "-->\n"
               "<method name='foo'>\n"
               "<!--\n"
               "And me!\n"
               "-->\n"
               "<arg name='bar' type='s'/>\n"
               "<arg name='no-comment' type='s'/>\n"
               "</method>\n"
               "</interface></node>\n")

        (parser, interfaces, _) = _test_parser(xml)
        interface = interfaces.get('I.I')
        self.assertIsNotNone(interface)
        self.assertEqual(interface.line_number, 5)
        self.assertEqual(interface.comment_line_number, 2)
        meth = interface.methods.get('foo')
        self.assertIsNotNone(meth)
        self.assertEqual(meth.line_number, 9)
        self.assertEqual(meth.comment_line_number, 6)
        self.assertEquals(len(meth.arguments), 2)
        arg = meth.arguments[0]
        self.assertEqual(arg.line_number, 13)
        self.assertEqual(arg.comment_line_number, 10)
        arg = meth.arguments[1]
        self.assertEqual(arg.line_number, 14)
        self.assertEqual(arg.comment_line_number, -1)

    def test_doc_annotations(self):
        xml = ("<node xmlns:tp='"
               "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
               "' xmlns:doc='"
               "http://www.freedesktop.org/dbus/1.0/doc.dtd'>"
               "<interface name='I.I'>"
               "<annotation name='org.gtk.GDBus.DocString' value='bla'/>"
               "</interface></node>")
        (parser, interfaces, _) = _test_parser(xml)
        interface = interfaces.get('I.I')
        self.assertIsNotNone(interface)
        self.assertEquals(interface.comment, "bla")

    def test_multiline_comments(self):
        xml = ("<node xmlns:tp='"
               "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
               "' xmlns:doc='"
               "http://www.freedesktop.org/dbus/1.0/doc.dtd'>"
               "<!--"
               "    Please consider that\n"
               "    multiline comment"
               "-->"
               "<interface name='I.I'>"
               "</interface></node>")

        (parser, interfaces, _) = _test_parser(xml)
        interface = interfaces.get('I.I')
        self.assertIsNotNone(interface)
        self.assertEquals(interface.comment,
                          "    Please consider that\n"
                          "    multiline comment")

    def test_ignored_comments(self):
        xml = ("<node xmlns:tp='"
               "http://telepathy.freedesktop.org/wiki/DbusSpec#extensions-v0"
               "' xmlns:doc='"
               "http://www.freedesktop.org/dbus/1.0/doc.dtd'>"
               "<!--"
               "Please ignore that comment"
               "-->"
               "<tp:copyright>"
               "</tp:copyright>"
               "<interface name='I.I'>"
               "</interface></node>")

        (parser, interfaces, _) = _test_parser(xml)
        interface = interfaces.get('I.I')
        self.assertIsNotNone(interface)
        self.assertIsNone(interface.comment)

    def test_root_node(self):
        self.assertParse("""<node name="/" />""")

    def test_dbus_spec_example(self):
        """
        Test parsing the example from the D-Bus Specification:

            http://dbus.freedesktop.org/doc/dbus-specification.html#introspection-format
        """
        self.assertParse("""
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
 "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node name="/com/example/sample_object">
  <interface name="com.example.SampleInterface">
    <method name="Frobate">
      <arg name="foo" type="i" direction="in"/>
      <arg name="bar" type="s" direction="out"/>
      <arg name="baz" type="a{us}" direction="out"/>
      <annotation name="org.freedesktop.DBus.Deprecated" value="true"/>
    </method>
    <method name="Bazify">
      <arg name="bar" type="(iiu)" direction="in"/>
      <arg name="bar" type="v" direction="out"/>
    </method>
    <method name="Mogrify">
      <arg name="bar" type="(iiav)" direction="in"/>
    </method>
    <signal name="Changed">
      <arg name="new_value" type="b"/>
    </signal>
    <property name="Bar" type="y" access="readwrite"/>
  </interface>
  <node name="child_of_sample_object"/>
  <node name="another_child_of_sample_object"/>
</node>""")


class TestParserOutputCodes(unittest.TestCase):
    """Test the output codes from InterfaceParser."""

    def test_unique(self):
        codes = InterfaceParser.get_output_codes()
        self.assertEqual(len(codes), len(set(codes)))

    def test_non_empty(self):
        codes = InterfaceParser.get_output_codes()
        self.assertNotEqual(codes, [])


# pylint: disable=too-many-public-methods
class TestParserRecovery(unittest.TestCase):
    """
    Test recovery from parser errors in InterfaceParser to detect subsequent
    errors.
    """

    # pylint: disable=invalid-name
    def assertOutput(self, xml, partial_output):  # noqa
        (parser, interfaces, filename) = _test_parser(xml)
        self.assertEqual(interfaces, None)
        actual_output = \
            [(filename, 'parser', i[0], i[1]) for i in partial_output]
        self.assertEqual(parser.get_output(), actual_output)
        return interfaces

    def test_annotation_interface(self):
        """Test recovery from invalid annotations in an interface."""
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<annotation/><method/>"
            "</interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in annotation.'),
                ('missing-attribute',
                 'Missing required attribute ‘name’ in method.'),
            ])

    def test_annotation_method(self):
        """Test recovery from invalid annotations in a method."""
        self.assertOutput(
            "<node><interface name='I.I'><method name='M'>"
            "<annotation/><arg/>"
            "</method></interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in annotation.'),
                ('missing-attribute',
                 'Missing required attribute ‘type’ in arg.'),
            ])

    def test_annotation_signal(self):
        """Test recovery from invalid annotations in a signal."""
        self.assertOutput(
            "<node><interface name='I.I'><signal name='S'>"
            "<annotation/><arg/>"
            "</signal></interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in annotation.'),
                ('missing-attribute',
                 'Missing required attribute ‘type’ in arg.'),
            ])

    def test_annotation_property(self):
        """Test recovery from invalid annotations in a property."""
        self.assertOutput(
            "<node><interface name='I.I'>"
            "<property name='P' type='s' access='read'>"
            "<annotation/><badnode/>"
            "</property>"
            "</interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in annotation.'),
                ('unknown-node',
                 'Unknown node ‘badnode’ in property ‘I.I.P’.'),
            ])

    def test_annotation_arg(self):
        """Test recovery from invalid annotations in an arg."""
        self.assertOutput(
            "<node><interface name='I.I'><method name='M'><arg type='s'>"
            "<annotation/><badnode/>"
            "</arg></method></interface></node>", [
                ('missing-attribute',
                 'Missing required attribute ‘name’ in annotation.'),
                ('unknown-node',
                 'Unknown node ‘badnode’ in argument ‘0 of method ‘I.I.M’’.'),
            ])


if __name__ == '__main__':
    # Run test suite
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright © 2015 Collabora Ltd.
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrapper around dbus-interface-diff to integrate it with a VCS

This implements an API signature database in the project’s version control
system so that all users of the VCS can do API compatibility between all past
signed releases. Currently, only git is supported.

Requirements:
 • Support out-of-tree builds, where srcdir ≠ builddir.
 • Support bare repositories, or where GIT_WORK_TREE ≠ .git.
 • Exit with status 0 if GIT_DIR does not exist; support the use case of
   running `make distcheck` inside an extracted release tarball.
 • Support both tag-before-dist and tag-after-dist workflows.
"""

import argparse
from contextlib import contextmanager
import os
import pipes
import shlex
import shutil
import subprocess
import sys
import tempfile


@contextmanager
def named_pipe():
    """Create and cleanup a named pipe in a temporary directory."""
    dirname = tempfile.mkdtemp()
    try:
        path = os.path.join(dirname, 'fifo')
        os.mkfifo(path)
        yield path
    finally:
        shutil.rmtree(dirname)


def _git_command(args, command):
    """Build a git command line with standard arguments."""
    out = [args.git]
    if args.git_dir != '':
        out += ['--git-dir', args.git_dir]
    if args.git_work_tree != '':
        out += ['--work-tree', args.git_work_tree]

    return out + [command]


def _format_command(args):
    """Local wrapper for pipes.quote() with support for Python 2.7."""
    try:
        # pylint: disable=no-member
        return ' '.join(shlex.quote(a) for a in args)
    except AttributeError:
        try:
            # pylint: disable=no-member
            return ' '.join(pipes.quote(a) for a in args)
        except AttributeError:
            # Give up
            return ' '.join(args)


def _get_contents_of_file(args, tag, api_xml_file):
    """Get the git object ID of api_xml_file in the tag revision."""
    rev = subprocess.check_output(_git_command(args, 'rev-parse') +
                                  ['--verify', '--quiet',
                                   '%s^{tag}:%s' % (tag, api_xml_file)])
    return rev.strip().decode('utf-8')


def _set_notes_for_ref(args, tag, api_xml_basename, notes):
    """Store the notes object ID as api_xml_basename in the tag revision."""
    with open(os.devnull, 'w') as dev_null:
        subprocess.check_output(_git_command(args, 'notes') +
                                ['--ref',
                                 'refs/%s/%s' %
                                 (args.dbus_api_git_refs, api_xml_basename),
                                 'add', '-C', notes, tag],
                                stderr=dev_null)


def _notes_exist_for_ref(args, tag, api_xml_basename):
    """Check whether notes have already been stored for this file and tag."""
    with open(os.devnull, 'w') as dev_null:
        status = subprocess.call(_git_command(args, 'notes') +
                                 ['--ref',
                                  'refs/%s/%s' %
                                  (args.dbus_api_git_refs, api_xml_basename),
                                  'show', tag],
                                 stdout=dev_null,
                                 stderr=dev_null)
        return status == 0


def _get_notes_filename_for_head(args, api_xml_basename):
    """Get the filename of api_xml_basename in the git work tree."""
    filename = subprocess.check_output(_git_command(args, 'ls-files') +
                                       ['--full-name',
                                        '*/%s' % api_xml_basename])
    filename = filename.strip().decode('utf-8')

    # Resolve the relative path against the work tree.
    if args.git_work_tree != '':
        filename = os.path.join(args.git_work_tree, filename)

    return filename


def _fetch_notes(args):
    """Fetch the latest API signature database from the remote."""
    subprocess.check_output(_git_command(args, 'fetch') +
                            [args.git_remote_origin,
                             'refs/%s/*:refs/%s/*' %
                             (args.dbus_api_git_refs, args.dbus_api_git_refs)])


def _push_notes(args):
    """Push the local API signature database to the remote."""
    command = _git_command(args, 'push')
    command += [args.git_remote_origin,
                'refs/' + args.dbus_api_git_refs + '/*']

    if not args.no_push:
        subprocess.check_output(command)
    else:
        sys.stdout.write('Run this command to push the API signature '
                         'database:\n'
                         '   %s\n' % _format_command(command))


def _is_release(args, ref):
    """Check whether ref identifies a signed tag."""
    with open(os.devnull, 'w') as dev_null:
        code = subprocess.call(_git_command(args, 'rev-parse') +
                               ['--verify', ref],
                               stdout=dev_null, stderr=dev_null)
    return code == 0


def _get_latest_release(args):
    """Get the name of the latest signed tag."""
    tag_list = subprocess.check_output(_git_command(args, 'rev-list') +
                                       ['--tags', '--max-count=1'])
    tag_list = tag_list.strip().decode('utf-8').split('\n')
    latest_tag = subprocess.check_output(_git_command(args, 'describe') +
                                         ['--tags'] + tag_list)

    return latest_tag.strip().decode('utf-8')


def command_dist(args):
    """Store the current API signature against the latest signed tag."""
    # Get the latest git tag
    try:
        latest_tag = _get_latest_release(args)
    except subprocess.CalledProcessError:
        sys.stderr.write('error: Failed to find latest git tag: %s.')
        return 1

    # Store notes for each API file
    for api_xml_file in args.dbus_api_xml_files:
        try:
            api_xml_basename = os.path.basename(api_xml_file)

            # Do notes already exist for this file and tag?
            if args.ignore_existing and \
               _notes_exist_for_ref(args, latest_tag, api_xml_basename):
                sys.stdout.write('%s: Ignored XML file ‘%s’; already has '
                                 'a note\n' %
                                 (latest_tag, api_xml_basename))
                continue

            notes = _get_contents_of_file(args, latest_tag, api_xml_file)
            subprocess.check_output(_git_command(args, 'notes') +
                                    ['--ref',
                                     'refs/%s/%s' %
                                     (args.dbus_api_git_refs,
                                      api_xml_basename),
                                     'add', '-C', notes, latest_tag])

            sys.stdout.write('%s: Added note ‘%s’ for XML file ‘%s’\n' %
                             (latest_tag, notes, api_xml_basename))
        except subprocess.CalledProcessError:
            sys.stderr.write('error: Failed to store notes for API file '
                             '‘%s’ and git tag ‘%s’.\n' %
                             (api_xml_file, latest_tag))
            return 1

    # Push to the remote
    try:
        _push_notes(args)
    except subprocess.CalledProcessError:
        sys.stderr.write('error: Failed to push notes to remote ‘%s’.\n' %
                         args.git_remote_origin)
        return 1

    return 0


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def command_check(args):
    """
    Check for API differences between two tags.

    If old_ref is not specified, it defaults to the latest signed tag. If
    new_ref is not specified, it defaults to the git work tree.
    """
    if args.old_ref != '' and not _is_release(args, args.old_ref):
        sys.stderr.write('error: Invalid --old-ref ‘%s’\n' % args.old_ref)
        return 1

    if args.new_ref != '' and not _is_release(args, args.new_ref):
        sys.stderr.write('error: Invalid --new-ref ‘%s’\n' % args.new_ref)
        return 1

    try:
        _fetch_notes(args)
    except subprocess.CalledProcessError:
        # Continue anyway
        sys.stderr.write('error: Failed to fetch latest refs.\n')

    old_ref = args.old_ref
    new_ref = args.new_ref

    if old_ref == '':
        # Get the latest git tag
        try:
            old_ref = _get_latest_release(args)
        except subprocess.CalledProcessError:
            sys.stderr.write('error: Failed to find latest git tag.\n')
            return 1

    try:
        refs = subprocess.check_output(_git_command(args, 'for-each-ref') +
                                       ['--format=%(refname)',
                                        'refs/%s' % args.dbus_api_git_refs])
        refs = refs.strip().decode('utf-8').split('\n')
    except subprocess.CalledProcessError:
        sys.stderr.write('error: Failed to get ref list.\n')
        return 1

    retval = 0

    for note_ref in refs:
        api_xml_basename = os.path.basename(note_ref)

        if args.silent:
            sys.stdout.write(' DIFF      %s\n' % api_xml_basename)
        else:
            sys.stdout.write('Comparing %s\n' % api_xml_basename)

        with named_pipe() as old_pipe_path, named_pipe() as new_pipe_path:
            old_notes_filename = old_pipe_path

            if new_ref == '':
                new_notes_filename = \
                    _get_notes_filename_for_head(args, api_xml_basename)
            else:
                new_notes_filename = new_pipe_path

            diff_command = ['dbus-interface-diff',
                            '--warnings', args.warnings,
                            '--file-display-name', api_xml_basename]
            if args.fatal_warnings:
                diff_command += ['--fatal-warnings']
            diff_command += [old_notes_filename, new_notes_filename]
            old_notes_command = _git_command(args, 'notes') + [
                '--ref',
                'refs/%s/%s' % (args.dbus_api_git_refs, api_xml_basename),
                'show', old_ref,
            ]
            new_notes_command = _git_command(args, 'notes') + [
                '--ref',
                'refs/%s/%s' % (args.dbus_api_git_refs, api_xml_basename),
                'show', new_ref,
            ]

            diff_proc = subprocess.Popen(diff_command)

            with open(old_pipe_path, 'wb') as old_pipe, \
                    open(os.devnull, 'w') as dev_null:
                old_notes_proc = subprocess.Popen(old_notes_command,
                                                  stdout=old_pipe,
                                                  stderr=dev_null)

                if new_ref == '':
                    new_notes_proc = None

                    # Debug output. Roughly equivalent to `set -v`.
                    if not args.silent:
                        ls_files_command = _git_command(args, 'ls-files') + [
                            '--full-name',
                            '*/%s' % api_xml_basename,
                        ]

                        if args.git_work_tree != '':
                            git_work_tree = args.git_work_tree + '/'
                        else:
                            git_work_tree = ''

                        sys.stdout.write('%s \\\n'
                                         '   <(%s) \\\n'
                                         '   %s`%s`\n' %
                                         (_format_command(diff_command[:-2]),
                                          _format_command(old_notes_command),
                                          git_work_tree,
                                          _format_command(ls_files_command)))
                else:
                    with open(new_pipe_path, 'wb') as new_pipe:
                        new_notes_proc = subprocess.Popen(new_notes_command,
                                                          stdout=new_pipe,
                                                          stderr=dev_null)

                    # Debug output. Roughly equivalent to `set -v`.
                    if not args.silent:
                        sys.stdout.write('%s \\\n'
                                         '   <(%s) \\\n'
                                         '   <(%s)\n' %
                                         (_format_command(diff_command[:-2]),
                                          _format_command(old_notes_command),
                                          _format_command(new_notes_command)))

                old_notes_proc.communicate()
                old_notes_proc.wait()

                if new_notes_proc is not None:
                    new_notes_proc.wait()

            diff_proc.wait()

            # Output the status from the first failure
            if retval == 0 and diff_proc.returncode != 0:
                retval = diff_proc.returncode

    return retval


def command_install(args):
    """Set up the API signature database for all existing signed tags."""
    try:
        tag_list = subprocess.check_output(_git_command(args, 'tag'))
        tag_list = tag_list.strip().decode('utf-8').split('\n')
    except subprocess.CalledProcessError:
        sys.stderr.write('error: Failed to get tag list.\n')
        return 1

    for tag in tag_list:
        outputted = False

        for api_xml_file in args.dbus_api_xml_files:
            api_xml_basename = os.path.basename(api_xml_file)

            try:
                notes = _get_contents_of_file(args, tag, api_xml_file)
            except subprocess.CalledProcessError:
                # Ignore it.
                notes = ''

            if notes == '':
                continue

            try:
                _set_notes_for_ref(args, tag, api_xml_basename, notes)

                sys.stdout.write('%s: Added note ‘%s’ for XML file ‘%s’\n' %
                                 (tag, notes, api_xml_basename))
                outputted = True
            except subprocess.CalledProcessError:
                # Ignore it
                continue

        if not outputted:
            sys.stdout.write('%s: Nothing to do\n' % tag)

    # Push the new refs
    try:
        _push_notes(args)
    except subprocess.CalledProcessError:
        sys.stderr.write('error: Failed to push notes to remote ‘%s’.\n' %
                         args.git_remote_origin)
        return 1

    return 0


def main():
    """Main helper implementation."""
    # Parse command line arguments.
    parser = argparse.ArgumentParser(
        description='Comparing D-Bus interface definitions')

    # Common arguments
    parser.add_argument('--silent', action='store_const', const=True,
                        default=False,
                        help='Silence all non-error output')
    parser.add_argument('--git', type=str, default='git', metavar='COMMAND',
                        help='Path to the git command, including extra '
                             'arguments')
    parser.add_argument('--git-dir', type=str, default='', metavar='PATH',
                        help='Path to the git directory in the project '
                             'checkout')
    parser.add_argument('--git-work-tree', type=str, default='',
                        metavar='PATH',
                        help='Path to the git work tree for the project')
    parser.add_argument('--git-remote', dest='git_remote_origin', type=str,
                        default='origin', metavar='REMOTE',
                        help='git remote to push notes to')
    parser.add_argument('--git-refs', dest='dbus_api_git_refs', type=str,
                        default='notes/dbus/api', metavar='REF-PATH',
                        help='Path beneath refs/ where the git notes will be'
                             ' stored containing the API signatures database')
    parser.add_argument('--no-push', action='store_const', const=True,
                        default=False,
                        help='Disable automatic pushing the API signature '
                             'database to a remote repository')

    subparsers = parser.add_subparsers()

    # dist command
    parser_dist = subparsers.add_parser('dist')
    parser_dist.add_argument('dbus_api_xml_files', metavar='API-FILE',
                             type=str, nargs='+',
                             help='D-Bus API XML file to check')
    parser_dist.add_argument('--ignore-existing', action='store_const',
                             const=True, default=False,
                             help='Ignore existing API signatures rather than '
                                  'erroring')
    parser_dist.set_defaults(func=command_dist)

    # check command
    parser_check = subparsers.add_parser('check')
    parser_check.add_argument('--diff-warnings', dest='warnings', type=str,
                              default='all',
                              help='Comma-separated list of warnings to '
                                   'enable when running dbus-interface-diff')
    parser_check.add_argument('--fatal-warnings', action='store_const',
                              const=True, default=False,
                              help='Treat all warnings as fatal')
    parser_check.add_argument('old_ref', metavar='OLD-REF',
                              type=str, nargs='?', default='',
                              help='Old ref to compare; or empty for the '
                                   'latest signed tag')
    parser_check.add_argument('new_ref', metavar='NEW-REF',
                              type=str, nargs='?', default='',
                              help='New ref to compare; or empty for HEAD')
    parser_check.set_defaults(func=command_check)

    # install command
    parser_install = subparsers.add_parser('install')
    parser_install.add_argument('dbus_api_xml_files', metavar='API-FILE',
                                type=str, nargs='+',
                                help='D-Bus API XML file to install')
    parser_install.set_defaults(func=command_install)

    args = parser.parse_args()

    # Bail early if the .git directory does not exist, since that's our data
    # store.
    git_dir = args.git_dir
    if git_dir == '':
        git_dir = os.path.join(args.git_work_tree, '.git')
    if not os.path.isdir(git_dir):
        sys.stderr.write('error: Could not find git directory ‘%s’. '
                         'Skipping.\n' % git_dir)
        return 0

    return args.func(args)


if __name__ == '__main__':
    main()
